/*
 * Copyright (C) 2004 Claudio "KLaN" Ciccani <klan82@cheapnet.it>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *  video_out_dfb: unofficial xine video output driver using DirectFB
 *
 *
 */


static struct
{
     int16_t vr[4]; /* v red coeff.   (offset  0) */
     int16_t vg[4]; /* v green coeff. (offset  8) */
     int16_t ug[4]; /* u green coeff. (offset 16) */
     int16_t ub[4]; /* u blue coeff.  (offset 24) */
     int16_t br[4]; /* brightness     (offset 32) */
     int16_t cn[4]; /* contrast       (offset 40) */
     int16_t st[4]; /* saturation     (offset 48) */

} __attribute__ ((aligned( 8 ))) mmx_table;


#define MMX_GEN_VR() \
{\
     mmx_table.vr[1] = mmx_table.vr[0] = ((V_RED_FACTOR >> 1) * saturation) >> 7;\
     mmx_table.vr[3] = mmx_table.vr[2] = mmx_table.vr[1];\
}

#define MMX_GEN_VG() \
{\
     mmx_table.vg[1] = mmx_table.vg[0] = ((V_GREEN_FACTOR >> 1) * saturation) >> 7;\
     mmx_table.vg[3] = mmx_table.vg[2] = mmx_table.vg[1];\
}

#define MMX_GEN_UG() \
{\
     mmx_table.ug[1] = mmx_table.ug[0] = ((U_GREEN_FACTOR >> 1) * saturation) >> 7;\
     mmx_table.ug[3] = mmx_table.ug[2] = mmx_table.ug[1];\
}

#define MMX_GEN_UB() \
{\
     mmx_table.ub[1] = mmx_table.ub[0] = ((U_BLUE_FACTOR >> 1) * saturation) >> 7;\
     mmx_table.ub[3] = mmx_table.ub[2] = mmx_table.ub[1];\
}

#define MMX_GEN_BR() \
{\
     mmx_table.br[1] = mmx_table.br[0] = brightness;\
     mmx_table.br[3] = mmx_table.br[2] = mmx_table.br[1];\
}

#define MMX_GEN_CN() \
{\
     mmx_table.cn[1] = mmx_table.cn[0] = (contrast * 0x807f) >> 8;\
     mmx_table.cn[3] = mmx_table.cn[2] = mmx_table.cn[1];\
}

#define MMX_GEN_ST() \
{\
     mmx_table.st[1] = mmx_table.st[0] = (saturation * 0x807f) >> 8;\
     mmx_table.st[3] = mmx_table.st[2] = mmx_table.st[1];\
}




#define PACCEL  mmx


static const uint64_t chroma_sub[] = {0x0080008000800080ull};

static const uint64_t lmask[]      = {0x000000ff000000ffull};

static const uint64_t wmask[]      = {0x00ff00ff00ff00ffull};

static const uint64_t amask[]      = {0xff00ff00ff00ff00ull};

static const uint64_t b3_mask[]    = {0xe0e0e0e0e0e0e0e0ull};

static const uint64_t b5_mask[]    = {0xf8f8f8f8f8f8f8f8ull};

static const uint64_t b6_mask[]    = {0xfcfcfcfcfcfcfcfcull};


/******************************** Begin YUY2 callbacks **********************************/

static
DFB_PFUNCTION( yuy2, yuy2 )
{
     uint8_t *src = frame->vo_frame.base[0];
     int      n   = (frame->width.cur * frame->height.cur) >> 2;

     __asm__ __volatile__(

          "movq      (wmask), %%mm2\n\t" /* mm2 = [0 0xff 0 0xff 0 0xff 0 0xff] */
          "psllw     $8, %%mm2\n\t" /* mm2 = [0xff 0 0xff 0 0xff 0 0xff 0] */
          "movq      (chroma_sub), %%mm3\n\t" /* mm3 = [0 128 0 128 0 128 0 128] */
          "movq      (mmx_table + 32), %%mm4\n\t" /* mm4 = brightness */
          "movq      (mmx_table + 40), %%mm5\n\t" /* mm5 = contrast */
          "movq      (mmx_table + 48), %%mm6\n\t" /* mm6 = saturation */
          "pxor      %%mm7, %%mm7\n\t"
          ".align 16\n"
          "1:\n\t"
          "movq      (%1), %%mm0\n\t" /* mm0 = [v2 y3 u2 y2 v0 y1 u0 y0] */
          "movq      %%mm0, %%mm1\n\t"
          "pand      (wmask), %%mm1\n\t" /* mm1 = [0 y3 0 y2 0 y1 0 y0] */
          "testl     $3, %3\n\t"
          "jz        2f\n\t"
          "paddw     %%mm4, %%mm1\n\t" /* y + brightness */
          "psllw     $2, %%mm1\n\t" /* y << 2 */
          "pmulhw    %%mm5, %%mm1\n\t" /* y * contrast */
          "packuswb  %%mm1, %%mm1\n\t" /* mm1 = [y3 y2 y1 y0 y3 y2 y1 y0] */
          "punpcklbw %%mm7, %%mm1\n\t" /* mm1 = [0 y3 0 y2 0 y1 0 y0] */
          ".align 8\n"
          "2:\n\t"
          "pand      %%mm2, %%mm0\n\t" /* mm0 = [v2 0 u2 0 v0 0 u0 0] */
          "testl     $4, %3\n\t"
          "jz        3f\n\t"
          "psrlw     $8, %%mm0\n\t" /* mm0 = [0 v2 0 u2 0 v0 0 u0] */
          "psubw     %%mm3, %%mm0\n\t" /* (u, v) - 128 */
          "psllw     $2, %%mm0\n\t" /* (u, v) << 2 */
          "pmulhw    %%mm6, %%mm0\n\t" /* (u, v) * saturation */
          "paddw     %%mm3, %%mm0\n\t" /* (u, v) + 128 */ 
          "packuswb  %%mm0, %%mm0\n\t" /* mm0 = [v2 u2 v0 u0 v2 u2 v0 u0] */
          "punpcklbw %%mm7, %%mm0\n\t" /* mm0 = [0 v2 0 u2 0 v0 0 u0] */
          "psllw     $8, %%mm0\n\t" /* mm0 = [v2 0 u2 0 v0 0 u0 0] */
          ".align 8\n"
          "3:\n\t"
          "por       %%mm1, %%mm0\n\t" /* mm0 = [v2 y3 u2 y2 v0 y1 u0 y0] */
          "movq      %%mm0, (%0)\n\t"
          "addl      $8, %0\n\t"
          "addl      $8, %1\n\t"
          "decl      %2\n\t"
          "jnz       1b\n\t"
          "emms"
          :: "r" (dst), "r" (src), "q" (n),
             "q" (this->mixer.modified)
          : "memory");
}


static
DFB_PFUNCTION( yuy2, uyvy )
{
     uint8_t *src = frame->vo_frame.base[0];
     int      n   = (frame->width.cur * frame->height.cur) >> 2;

     __asm__ __volatile__(

          "movq      (chroma_sub), %%mm3\n\t" /* mm3 = [0 128 0 128 0 128 0 128] */
          "movq      (mmx_table + 32), %%mm4\n\t" /* mm4 = brightness */
          "movq      (mmx_table + 40), %%mm5\n\t" /* mm5 = contrast */
          "movq      (mmx_table + 48), %%mm6\n\t" /* mm6 = saturation */
          "pxor      %%mm7, %%mm7\n\t"
          ".align 16\n"
          "1:\n\t"
          "movq      (%1), %%mm0\n\t" /* mm0 = [v2 y3 u2 y2 v0 y1 u0 y0] */
          "movq      %%mm0, %%mm1\n\t"
          "pand      (wmask), %%mm1\n\t" /* mm1 = [0 y3 0 y2 0 y1 0 y0] */
          "psrlw     $8, %%mm0\n\t" /* mm0 = [0 v2 0 u2 0 v0 0 u0] */
          "testl     $3, %3\n\t"
          "jz        2f\n\t"
          "paddw     %%mm4, %%mm1\n\t" /* y + brightness */
          "psllw     $2, %%mm1\n\t" /* y << 2 */
          "pmulhw    %%mm5, %%mm1\n\t" /* y * contrast */
          "packuswb  %%mm1, %%mm1\n\t" /* mm1 = [y3 y2 y1 y0 y3 y2 y1 y0] */
          "punpcklbw %%mm7, %%mm1\n\t" /* mm1 = [0 y3 0 y2 0 y1 0 y0] */
          ".align 8\n"
          "2:\n\t"
          "psllw     $8, %%mm1\n\t" /* mm1 = [y3 0 y2 0 y1 0 y0 0] */
          "testl     $4, %3\n\t"
          "jz        3f\n\t"
          "psubw     %%mm3, %%mm0\n\t" /* (u, v) - 128 */
          "psllw     $2, %%mm0\n\t" /* (u, v) << 2 */
          "pmulhw    %%mm6, %%mm0\n\t" /* (u, v) * saturation */
          "paddw     %%mm3, %%mm0\n\t" /* (u, v) + 128 */
          "packuswb  %%mm0, %%mm0\n\t" /* mm0 = [v2 u2 v0 u0 v2 u2 v0 u0] */
          "punpcklbw %%mm7, %%mm0\n\t" /* mm0 = [0 v2 0 u2 0 v0 0 u0] */
          ".align 8\n"
          "3:\n\t"
          "por       %%mm1, %%mm0\n\t" /* mm0 = [y3 v2 y2 u2 y1 v0 y0 u0] */
          "movq      %%mm0, (%0)\n\t"
          "addl      $8, %0\n\t"
          "addl      $8, %1\n\t"
          "decl      %2\n\t"
          "jnz       1b\n\t"
          "emms"

          :: "r" (dst), "r" (src), "q" (n),
             "q" (this->mixer.modified)
          : "memory");
}


static
DFB_PFUNCTION( yuy2, yv12 )
{
     uint8_t *src   = frame->vo_frame.base[0];
     uint8_t *dsty  = dst;
     uint8_t *dstu;
     uint8_t *dstv;
     int      n     = frame->height.cur >> 1;

     if (frame->dstfmt.cur == DSPF_YV12)
     {
          dstv = (uint8_t*) dsty + (pitch * frame->height.cur);
          dstu = (uint8_t*) dstv + ((pitch * frame->height.cur) >> 2);
     } else
     {
          dstu = (uint8_t*) dsty + (pitch * frame->height.cur);
          dstv = (uint8_t*) dstu + ((pitch * frame->height.cur) >> 2);
     }

     __asm__ __volatile__(

          "xorl      %%ecx, %%ecx\n\t"
          ".align 16\n"
          "1:\n\t"
          "movq      (%0), %%mm0\n\t" /* mm0 = [v02 y03 u02 y02 v00 y01 u00 y00] */
          "movq      %%mm0, %%mm2\n\t" /* mm2 = [v02 y03 u02 y02 v00 y01 u00 y00] */
          "pand      (wmask), %%mm2\n\t" /* mm2 = [0 y03 0 y02 0 y01 0 y00] */
          "psrlw     $8, %%mm0\n\t" /* mm0 = [0 v02 0 u02 0 v00 0 u00] */
          "movq      8(%0), %%mm1\n\t" /* mm1 = [v06 y07 u06 y06 v04 y05 u04 y04] */
          "movq      %%mm1, %%mm3\n\t" /* mm3 = [v06 y07 u06 y06 v04 y05 u04 y04] */
          "pand      (wmask), %%mm3\n\t" /* mm3 = [0 y07 0 y06 0 y05 0 y04] */
          "psrlw     $8, %%mm1\n\t" /* mm1 = [0 v06 0 u06 0 v04 0 u04] */
          "testl     $3, %6\n\t"
          "jz        2f\n\t"
          "paddw     (mmx_table + 32), %%mm2\n\t" /* y + brightness */
          "psllw     $2, %%mm2\n\t" /* y << 2 */
          "pmulhw    (mmx_table + 40), %%mm2\n\t" /* y * contrast */
          "paddw     (mmx_table + 32), %%mm3\n\t" /* y + brightness */
          "psllw     $2, %%mm3\n\t" /* y << 2 */
          "pmulhw    (mmx_table + 40), %%mm3\n\t" /* y * contrast */
          ".align 8\n"
          "2:\n\t"
          "packuswb  %%mm3, %%mm2\n\t" /* mm2 = [y07 y06 y05 y04 y03 y02 y01 y00] */
          "movl      %1, %%eax\n\t" /* eax = dsty */
          "movq      %%mm2, (%%eax, %%ecx, 2)\n\t"
          "movq      (%0, %4), %%mm2\n\t" /* mm2 = [v12 y13 u12 y12 v10 y11 u10 y10] */
          "movq      %%mm2, %%mm4\n\t" /* mm4 = [v12 y13 u12 y12 v10 y11 u10 y10] */
          "pand      (wmask), %%mm4\n\t" /* mm4 = [0 y13 0 y12 0 y11 0 y10] */
          "psrlw     $8, %%mm2\n\t" /* mm2 = [0 v12 0 u12 0 v10 0 u10] */
          "movq      8(%0, %4), %%mm3\n\t" /* mm3 = [v16 y17 u16 y16 v14 y15 u14 y14] */
          "movq      %%mm3, %%mm5\n\t" /* mm5 = [v16 y17 u16 y16 v14 y15 u14 y14] */
          "pand      (wmask), %%mm5\n\t" /* mm5 = [0 y17 0 y16 0 y15 0 y14] */
          "psrlw     $8, %%mm3\n\t" /* mm3 = [0 v16 0 u16 0 v14 0 u14] */
          "testl     $3, %6\n\t"
          "jz        3f\n\t"
          "paddw     (mmx_table + 32), %%mm4\n\t" /* y + brightness */
          "psllw     $2, %%mm4\n\t" /* y << 2 */
          "pmulhw    (mmx_table + 40), %%mm4\n\t" /* y * contrast */
          "paddw     (mmx_table + 32), %%mm5\n\t" /* y + brightness */
          "psllw     $2, %%mm5\n\t" /* y << 2 */
          "pmulhw    (mmx_table + 40), %%mm5\n\t" /* y * contrast */
          ".align 8\n"
          "3:\n\t"
          "packuswb  %%mm5, %%mm4\n\t" /* mm4 = [y17 y16 y15 y14 y13 y12 y11 y10] */
          "addl      %5, %%eax\n\t" /* eax = dsty + pitch */
          "movq      %%mm4, (%%eax, %%ecx, 2)\n\t"
          "paddw     %%mm2, %%mm0\n\t"
          "psrlw     $1, %%mm0\n\t" /* mm0 = [0 v1 0 u1 0 v0 0 u0] */
          "paddw     %%mm3, %%mm1\n\t"
          "psrlw     $1, %%mm1\n\t" /* mm1 = [0 v3 0 u3 0 v2 0 u2] */
          "testl     $4, %6\n\t"
          "jz        4f\n\t"
          "psubw     (chroma_sub), %%mm0\n\t"/* (v1,u1,v0,u0) - 128 */
          "psllw     $2, %%mm0\n\t" /* (v1,u1,v0,u0) << 2 */
          "pmulhw    (mmx_table + 48), %%mm0\n\t" /* (v1,u1,v0,u0) * saturation */
          "paddw     (chroma_sub), %%mm0\n\t" /* (v1,u1,v0,u0) - 128 */
          "packuswb  %%mm0, %%mm0\n\t" /* mm0 = [v1 u1 v0 u0 v1 u1 v0 u0] */
          "punpcklbw %%mm7, %%mm0\n\t" /* mm0 = [0 v1 0 u1 0 v0 0 u0] */
          "psubw     (chroma_sub), %%mm1\n\t" /* (v3,u3,v2,u2) - 128 */
          "psllw     $2, %%mm1\n\t" /* (v3,u3,v2,u2) << 2 */
          "pmulhw    (mmx_table + 48), %%mm1\n\t" /* (v3,u3,v2,u2) * saturation */
          "paddw     (chroma_sub), %%mm1\n\t" /* (v3,u3,v2,u2) + 128 */
          "packuswb  %%mm1, %%mm1\n\t" /* mm1 = [v3 u3 v2 u2 v3 u3 v2 u2] */
          "punpcklbw %%mm7, %%mm1\n\t" /* mm1 = [0 v3 0 u3 0 v2 0 u2] */
          ".align 8\n"
          "4:\n\t"
          "movq      %%mm0, %%mm2\n\t" /* mm2 = [0 v1 0 u1 0 v0 0 u0] */
          "punpcklbw %%mm1, %%mm0\n\t" /* mm0 = [0 0 v2 v0 0 0 u2 u0] */
          "punpckhbw %%mm1, %%mm2\n\t" /* mm2 = [0 0 v3 v1 0 0 u3 u1] */
          "movq      %%mm0, %%mm1\n\t" /* mm1 = [0 0 v2 v0 0 0 u2 u0] */
          "punpcklbw %%mm2, %%mm1\n\t" /* mm1 = [0 0 0 0 u3 u2 u1 u0] */
          "movl      %2, %%eax\n\t"
          "movd      %%mm1, (%%eax, %%ecx)\n\t"
          "punpckhbw %%mm2, %%mm0\n\t" /* mm0 = [0 0 0 0 v3 v2 v1 v0] */
          "movl      %3, %%eax\n\t"
          "movd      %%mm0, (%%eax, %%ecx)\n\t"
          "addl      $16, %0\n\t"
          "addl      $4, %%ecx\n\t"
          "movl      %5, %%eax\n\t"
          "shrl      $1, %%eax\n\t"
          "cmpl      %%eax, %%ecx\n\t"
          "jl        1b\n\t"
          "xorl      %%ecx, %%ecx\n\t"
          "addl      %4, %0\n\t" /* src += frame->vo_frame.pitches[0] */
          "addl      %4, %1\n\t" /* dsty += pitch * 2 */ 
          "addl      %%eax, %2\n\t" /* dstu += pitch / 2 */
          "addl      %%eax, %3\n\t" /* dstv += pitch / 2 */
          "decl      %7\n\t"
          "jnz       1b\n\t"
          "emms"

          :: "r" (src), "m" (dsty), "m" (dstu), "m" (dstv),
             "r" (frame->vo_frame.pitches[0]), "r" (pitch), 
             "r" (this->mixer.modified), "m" (n)
          : "eax", "ecx");
}


static
DFB_PFUNCTION( yuy2, rgb8 )
{
     uint8_t *src = frame->vo_frame.base[0];
     int      n   = (frame->width.cur * frame->height.cur) >> 2;

     __asm__ __volatile__(

          "pxor      %%mm7, %%mm7\n\t"
          ".align 16\n"
          "1:\n\t"
          "movq      (%1), %%mm0\n\t" /* mm1 = [v2 y3 u2 y2 v0 y1 u0 y0] */
          "movq      %%mm0, %%mm2\n\t" /* mm2 = [v2 y3 u2 y2 v0 y1 u0 y0] */
          "pand      (wmask), %%mm2\n\t" /* mm2 = [0 y3 0 y2 0 y1 0 y0] */
          "paddw     (mmx_table + 32), %%mm2\n\t" /* y + brightness */
          "psllw     $2, %%mm2\n\t" /* y << 2 */
          "pmulhw    (mmx_table + 40), %%mm2\n\t" /* y * contrast */
          "movq      %%mm0, %%mm1\n\t" /* mm1 = [v2 y3 u2 y2 v0 y1 u0 y0] */
          "psrld     $8, %%mm1\n\t" /* mm1 = [0 v2 y3 u2 0 v0 y1 u0] */
          "pand      (lmask), %%mm1\n\t" /* mm1 = [0 0 0 u2 0 0 0 u0] */
          "packssdw  %%mm1, %%mm1\n\t" /* mm1 = [0 u2 0 u0 0 u2 0 u0] */
          "punpcklwd %%mm1, %%mm1\n\t" /* mm1 = [0 u2 0 u2 0 u0 0 u0] */
          "psubw     (chroma_sub), %%mm1\n\t" /* u -= 128 */
          "psllw     $3, %%mm1\n\t" /* u << 3 */
          "psrld     $24, %%mm0\n\t" /* mm0 = [0 0 0 v2 0 0 0 v0] */
          "packssdw  %%mm0, %%mm0\n\t" /* mm0 = [0 v2 0 v0 0 v2 0 v0] */
          "punpcklwd %%mm0, %%mm0\n\t" /* mm0 = [0 v2 0 v2 0 v0 0 v0] */
          "psubw     (chroma_sub), %%mm0\n\t" /* v -= 128 */
          "psllw     $3, %%mm0\n\t" /* v << 3 */
          "movq      %%mm0, %%mm3\n\t"
          "pmulhw    (mmx_table), %%mm3\n\t" /* chroma_r */
          "paddw     %%mm2, %%mm3\n\t" /* mm3 = [r3 | r2 | r1 | r0] */
          "packuswb  %%mm3, %%mm3\n\t" /* mm3 = [r3 r2 r1 r0 r3 r2 r1 r0] */
          "pand      (b3_mask), %%mm3\n\t" /* red 8 */
          "punpcklbw %%mm7, %%mm3\n\t" /* mm3 = [0 r3 0 r2 0 r1 0 r0] */
          "movq      %%mm1, %%mm4\n\t"
          "pmulhw    (mmx_table + 24), %%mm4\n\t" /* chroma_b */
          "paddw     %%mm2, %%mm4\n\t" /* mm4 = [b3 | b2 | b1 | b0] */
          "packuswb  %%mm4, %%mm4\n\t" /* mm4 = [b3 b2 b1 b0 b3 b2 b1 b0] */
          "punpcklbw %%mm7, %%mm4\n\t" /* mm4 = [0 b3 0 b2 0 b1 0 b0] */
          "psrlw     $6, %%mm4\n\t" /* blue 8 */
          "por       %%mm4, %%mm3\n\t"
          "pmulhw    (mmx_table + 8), %%mm0\n\t"
          "pmulhw    (mmx_table + 16), %%mm1\n\t"
          "paddw     %%mm1, %%mm0\n\t" /* chroma_g */
          "psubw     %%mm0, %%mm2\n\t" /* mm2 = [g3 | g2 | g1 | g0] */
          "packuswb  %%mm2, %%mm2\n\t" /* mm2 = [g3 g2 g1 g0 g3 g2 g1 g0] */
          "pand      (b3_mask), %%mm2\n\t"
          "punpcklbw %%mm7, %%mm2\n\t" /* mm2 = [0 g3 0 g2 0 g1 0 g0] */
          "psrlw     $3, %%mm2\n\t" /* green 8 */
          "por       %%mm2, %%mm3\n\t"
          "packuswb  %%mm3, %%mm3\n\t"
          "movd      %%mm3, (%0)\n\t" /* out 4 pixels */
          "addl      $4, %0\n\t"
          "addl      $8, %1\n\t"
          "decl      %2\n\t"
          "jnz       1b\n\t"
          "emms"

          :: "r" (dst), "r" (src), "q" (n)
          : "memory");
}


static
DFB_PFUNCTION( yuy2, rgb15 )
{
     uint8_t *src = frame->vo_frame.base[0];
     int      n   = (frame->width.cur * frame->height.cur) >> 2;

     __asm__ __volatile__(

          "pxor      %%mm7, %%mm7\n\t"
          ".align 16\n"
          "1:\n\t"
          "movq      (%1), %%mm0\n\t" /* mm1 = [v2 y3 u2 y2 v0 y1 u0 y0] */
          "movq      %%mm0, %%mm2\n\t" /* mm2 = [v2 y3 u2 y2 v0 y1 u0 y0] */
          "pand      (wmask), %%mm2\n\t" /* mm2 = [0 y3 0 y2 0 y1 0 y0] */
          "paddw     (mmx_table + 32), %%mm2\n\t" /* y + brightness */
          "psllw     $2, %%mm2\n\t" /* y << 2 */
          "pmulhw    (mmx_table + 40), %%mm2\n\t" /* y * contrast */
          "movq      %%mm0, %%mm1\n\t" /* mm1 = [v2 y3 u2 y2 v0 y1 u0 y0] */
          "psrld     $8, %%mm1\n\t" /* mm1 = [0 v2 y3 u2 0 v0 y1 u0] */
          "pand      (lmask), %%mm1\n\t" /* mm1 = [0 0 0 u2 0 0 0 u0] */
          "packssdw  %%mm1, %%mm1\n\t" /* mm1 = [0 u2 0 u0 0 u2 0 u0] */
          "punpcklwd %%mm1, %%mm1\n\t" /* mm1 = [0 u2 0 u2 0 u0 0 u0] */
          "psubw     (chroma_sub), %%mm1\n\t" /* u -= 128 */
          "psllw     $3, %%mm1\n\t" /* u << 3 */
          "psrld     $24, %%mm0\n\t" /* mm0 = [0 0 0 v2 0 0 0 v0] */
          "packssdw  %%mm0, %%mm0\n\t" /* mm0 = [0 v2 0 v0 0 v2 0 v0] */
          "punpcklwd %%mm0, %%mm0\n\t" /* mm0 = [0 v2 0 v2 0 v0 0 v0] */
          "psubw     (chroma_sub), %%mm0\n\t" /* v -= 128 */
          "psllw     $3, %%mm0\n\t" /* v << 3 */
          "movq      %%mm0, %%mm3\n\t"
          "pmulhw    (mmx_table), %%mm3\n\t" /* chroma_r */
          "paddw     %%mm2, %%mm3\n\t" /* mm3 = [r3 | r2 | r1 | r0] */
          "packuswb  %%mm3, %%mm3\n\t" /* mm3 = [r3 r2 r1 r0 r3 r2 r1 r0] */
          "pand      (b5_mask), %%mm3\n\t" /* red 15 */
          "movq      %%mm1, %%mm4\n\t"
          "pmulhw    (mmx_table + 24), %%mm4\n\t" /* chroma_b */
          "paddw     %%mm2, %%mm4\n\t" /* mm4 = [b3 | b2 | b1 | b0] */
          "packuswb  %%mm4, %%mm4\n\t" /* mm4 = [b3 b2 b1 b0 b3 b2 b1 b0] */
          "pmulhw    (mmx_table + 8), %%mm0\n\t"
          "pmulhw    (mmx_table + 16), %%mm1\n\t"
          "paddw     %%mm1, %%mm0\n\t" /* chroma_g */
          "psubw     %%mm0, %%mm2\n\t" /* mm2 = [g3 | g2 | g1 | g0] */
          "packuswb  %%mm2, %%mm2\n\t" /* mm2 = [g3 g2 g1 g0 g3 g2 g1 g0] */
          "pand      (b5_mask), %%mm2\n\t" /* green 15 */
          "punpcklbw %%mm7, %%mm3\n\t" /* mm3 = [0 r3 0 r2 0 r1 0 r0] */
          "psllw     $7, %%mm3\n\t"
          "punpcklbw %%mm7, %%mm4\n\t" /* mm4 = [0 b3 0 b2 0 b1 0 b0] */
          "psrlw     $3, %%mm4\n\t" /* blue 15 */
          "por       %%mm4, %%mm3\n\t"
          "punpcklbw %%mm7, %%mm2\n\t" /* mm2 = [0 g3 0 g2 0 g1 0 g0] */
          "psllw     $2, %%mm2\n\t" /* green 15 */
          "por       %%mm2, %%mm3\n\t" /* rgb15 */
          "movq      %%mm3, (%0)\n\t" /* out 4 pixles */
          "addl      $8, %0\n\t"
          "addl      $8, %1\n\t"
          "decl      %2\n\t"
          "jnz       1b\n\t"
          "emms"

          :: "r" (dst), "r" (src), "q" (n)
          : "memory");
}


static
DFB_PFUNCTION( yuy2, rgb16 )
{
     uint8_t *src = frame->vo_frame.base[0];
     int      n   = (frame->width.cur * frame->height.cur) >> 2;

     __asm__ __volatile__(

          "pxor      %%mm7, %%mm7\n\t"
          ".align 16\n"
          "1:\n\t"
          "movq      (%1), %%mm0\n\t" /* mm1 = [v2 y3 u2 y2 v0 y1 u0 y0] */
          "movq      %%mm0, %%mm2\n\t" /* mm2 = [v2 y3 u2 y2 v0 y1 u0 y0] */
          "pand      (wmask), %%mm2\n\t" /* mm2 = [0 y3 0 y2 0 y1 0 y0] */
          "paddw     (mmx_table + 32), %%mm2\n\t" /* y + brightness */
          "psllw     $2, %%mm2\n\t" /* y << 2 */
          "pmulhw    (mmx_table + 40), %%mm2\n\t" /* y * contrast */
          "movq      %%mm0, %%mm1\n\t" /* mm1 = [v2 y3 u2 y2 v0 y1 u0 y0] */
          "psrld     $8, %%mm1\n\t" /* mm1 = [0 v2 y3 u2 0 v0 y1 u0] */
          "pand      (lmask), %%mm1\n\t" /* mm1 = [0 0 0 u2 0 0 0 u0] */
          "packssdw  %%mm1, %%mm1\n\t" /* mm1 = [0 u2 0 u0 0 u2 0 u0] */
          "punpcklwd %%mm1, %%mm1\n\t" /* mm1 = [0 u2 0 u2 0 u0 0 u0] */
          "psubw     (chroma_sub), %%mm1\n\t" /* u -= 128 */
          "psllw     $3, %%mm1\n\t" /* u << 3 */
          "psrld     $24, %%mm0\n\t" /* mm0 = [0 0 0 v2 0 0 0 v0] */
          "packssdw  %%mm0, %%mm0\n\t" /* mm0 = [0 v2 0 v0 0 v2 0 v0] */
          "punpcklwd %%mm0, %%mm0\n\t" /* mm0 = [0 v2 0 v2 0 v0 0 v0] */
          "psubw     (chroma_sub), %%mm0\n\t" /* v -= 128 */
          "psllw     $3, %%mm0\n\t" /* v << 3 */
          "movq      %%mm0, %%mm3\n\t"
          "pmulhw    (mmx_table), %%mm3\n\t" /* chroma_r */
          "paddw     %%mm2, %%mm3\n\t" /* mm3 = [r3 | r2 | r1 | r0] */
          "packuswb  %%mm3, %%mm3\n\t" /* mm3 = [r3 r2 r1 r0 r3 r2 r1 r0] */
          "pand      (b5_mask), %%mm3\n\t" /* red 16 */
          "movq      %%mm1, %%mm4\n\t"
          "pmulhw    (mmx_table + 24), %%mm4\n\t" /* chroma_b */
          "paddw     %%mm2, %%mm4\n\t" /* mm4 = [b3 | b2 | b1 | b0] */
          "packuswb  %%mm4, %%mm4\n\t" /* mm4 = [b3 b2 b1 b0 b3 b2 b1 b0] */
          "pmulhw    (mmx_table + 8), %%mm0\n\t"
          "pmulhw    (mmx_table + 16), %%mm1\n\t"
          "paddw     %%mm1, %%mm0\n\t" /* chroma_g */
          "psubw     %%mm0, %%mm2\n\t" /* mm2 = [g3 | g2 | g1 | g0] */
          "packuswb  %%mm2, %%mm2\n\t" /* mm2 = [g3 g2 g1 g0 g3 g2 g1 g0] */
          "pand      (b6_mask), %%mm2\n\t" /* green 16 */
          "punpcklbw %%mm7, %%mm3\n\t" /* mm3 = [0 r3 0 r2 0 r1 0 r0] */
          "psllw     $8, %%mm3\n\t"
          "punpcklbw %%mm7, %%mm4\n\t" /* mm4 = [0 b3 0 b2 0 b1 0 b0] */
          "psrlw     $3, %%mm4\n\t" /* blue 16 */
          "por       %%mm4, %%mm3\n\t"
          "punpcklbw %%mm7, %%mm2\n\t" /* mm2 = [0 g3 0 g2 0 g1 0 g0] */
          "psllw     $3, %%mm2\n\t"
          "por       %%mm2, %%mm3\n\t" /* rgb16 */
          "movq      %%mm3, (%0)\n\t" /* out 4 pixels */
          "addl      $8, %0\n\t"
          "addl      $8, %1\n\t"
          "decl      %2\n\t"
          "jnz       1b\n\t"
          "emms"

          :: "r" (dst), "r" (src), "q" (n)
          : "memory");
}


static
DFB_PFUNCTION( yuy2, rgb24 )
{
     uint8_t *src = frame->vo_frame.base[0];
     int      n   = (frame->width.cur * frame->height.cur) >> 2;

     __asm__ __volatile__(

          "pxor      %%mm7, %%mm7\n\t"
          ".align 16\n"
          "1:\n\t"
          "movq      (%1), %%mm0\n\t" /* mm1 = [v2 y3 u2 y2 v0 y1 u0 y0] */
          "movq      %%mm0, %%mm2\n\t" /* mm2 = [v2 y3 u2 y2 v0 y1 u0 y0] */
          "pand      (wmask), %%mm2\n\t" /* mm2 = [0 y3 0 y2 0 y1 0 y0] */
          "paddw     (mmx_table + 32), %%mm2\n\t" /* y + brightness */
          "psllw     $2, %%mm2\n\t" /* y << 2 */
          "pmulhw    (mmx_table + 40), %%mm2\n\t" /* y * contrast */
          "movq      %%mm0, %%mm1\n\t" /* mm1 = [v2 y3 u2 y2 v0 y1 u0 y0] */
          "psrld     $8, %%mm1\n\t" /* mm1 = [0 v2 y3 u2 0 v0 y1 u0] */
          "pand      (lmask), %%mm1\n\t" /* mm1 = [0 0 0 u2 0 0 0 u0] */
          "packssdw  %%mm1, %%mm1\n\t" /* mm1 = [0 u2 0 u0 0 u2 0 u0] */
          "punpcklwd %%mm1, %%mm1\n\t" /* mm1 = [0 u2 0 u2 0 u0 0 u0] */
          "psubw     (chroma_sub), %%mm1\n\t" /* u -= 128 */
          "psllw     $3, %%mm1\n\t" /* u << 3 */
          "psrld     $24, %%mm0\n\t" /* mm0 = [0 0 0 v2 0 0 0 v0] */
          "packssdw  %%mm0, %%mm0\n\t" /* mm0 = [0 v2 0 v0 0 v2 0 v0] */
          "punpcklwd %%mm0, %%mm0\n\t" /* mm0 = [0 v2 0 v2 0 v0 0 v0] */
          "psubw     (chroma_sub), %%mm0\n\t" /* v -= 128 */
          "psllw     $3, %%mm0\n\t" /* v << 3 */
          "movq      %%mm0, %%mm3\n\t"
          "pmulhw    (mmx_table), %%mm3\n\t" /* chroma_r */
          "paddw     %%mm2, %%mm3\n\t" /* mm3 = [r3 | r2 | r1 | r0] */
          "packuswb  %%mm3, %%mm3\n\t" /* mm3 = [r3 r2 r1 r0 r3 r2 r1 r0] */
          "movq      %%mm1, %%mm4\n\t"
          "pmulhw    (mmx_table + 24), %%mm4\n\t" /* chroma_b */
          "paddw     %%mm2, %%mm4\n\t" /* mm4 = [b3 | b2 | b1 | b0] */
          "packuswb  %%mm4, %%mm4\n\t" /* mm4 = [b3 b2 b1 b0 b3 b2 b1 b0] */
          "pmulhw    (mmx_table + 8), %%mm0\n\t"
          "pmulhw    (mmx_table + 16), %%mm1\n\t"
          "paddw     %%mm1, %%mm0\n\t" /* chroma_g */
          "psubw     %%mm0, %%mm2\n\t" /* mm2 = [g3 | g2 | g1 | g0] */
          "packuswb  %%mm2, %%mm2\n\t" /* mm2 = [g3 g2 g1 g0 g3 g2 g1 g0] */
          "movq      %%mm4, %%mm0\n\t" /* mm0 = [b3 b2 b1 b0 b3 b2 b1 b0] */
          "punpcklbw %%mm2, %%mm0\n\t" /* mm0 = [g3 b3 g2 b2 g1 b1 g0 b0] */
          "movq      %%mm0, %%mm1\n\t" /* mm1 = [g3 b3 g2 b2 g1 b1 g0 b0] */
          "punpcklbw %%mm3, %%mm4\n\t" /* mm4 = [r3 b3 r2 b2 r1 b1 r0 b0] */
          "psrld $8, %%mm4\n\t" /* mm4 = [0 r3 b3 r2 0 r1 b1 r0] */
          "punpcklwd %%mm4, %%mm0\n\t" /* mm0 = [0 r1 g1 b1 b1 r0 g0 b0] */
          "movq      %%mm0, %%mm6\n\t" /* mm6 = [0 r1 g1 b1 b1 r0 g0 b0] */
          "psrlq     $40, %%mm6\n\t" /* mm6 = [0 0 0 0 0 0 r1 g1] */
          "punpckhwd %%mm4, %%mm1\n\t" /* mm1 = [0 r3 g3 b3 b3 r2 g2 b2] */
          "punpcklwd %%mm1, %%mm6\n\t" /* mm6 = [b3 r2 0 0 g2 b2 r1 g1] */
          "punpckldq %%mm6, %%mm0\n\t" /* mm0 = [g2 b2 r1 g1 b1 r0 g0 b0] */
          "movq      %%mm0, (%0)\n\t" /* out 2 + 2/3 pixels */
          "psrld     $8, %%mm1\n\t" /* mm1 = [0 0 r3 g3 0 b3 r2 g2] */
          "psrld     $16, %%mm6\n\t" /* mm6 = [0 0 b3 r2 0 0 g2 b2] */
          "punpckhwd %%mm1, %%mm6\n\t" /* mm6 = [0 0 0 0 r3 g3 b3 r2] */
          "movd      %%mm6, 8(%0)\n\t" /* out 1/3 + 1 pixels */
          "addl      $12, %0\n\t"
          "addl      $8, %1\n\t"
          "decl      %2\n\t"
          "jnz       1b\n\t"
          "emms"

          :: "r" (dst), "r" (src), "q" (n)
          : "memory");
}


static
DFB_PFUNCTION( yuy2, rgb32 )
{
     uint8_t *src = frame->vo_frame.base[0];
     int      n   = (frame->width.cur * frame->height.cur) >> 2;

     __asm__ __volatile__(

          "pxor      %%mm7, %%mm7\n\t"
          ".align 16\n"
          "1:\n\t"
          "movq      (%1), %%mm0\n\t" /* mm1 = [v2 y3 u2 y2 v0 y1 u0 y0] */
          "movq      %%mm0, %%mm2\n\t" /* mm2 = [v2 y3 u2 y2 v0 y1 u0 y0] */
          "pand      (wmask), %%mm2\n\t" /* mm2 = [0 y3 0 y2 0 y1 0 y0] */
          "paddw     (mmx_table + 32), %%mm2\n\t" /* y + brightness */
          "psllw     $2, %%mm2\n\t" /* y << 2 */
          "pmulhw    (mmx_table + 40), %%mm2\n\t" /* y * contrast */
          "movq      %%mm0, %%mm1\n\t" /* mm1 = [v2 y3 u2 y2 v0 y1 u0 y0] */
          "psrld     $8, %%mm1\n\t" /* mm1 = [0 v2 y3 u2 0 v0 y1 u0] */
          "pand      (lmask), %%mm1\n\t" /* mm1 = [0 0 0 u2 0 0 0 u0] */
          "packssdw  %%mm1, %%mm1\n\t" /* mm1 = [0 u2 0 u0 0 u2 0 u0] */
          "punpcklwd %%mm1, %%mm1\n\t" /* mm1 = [0 u2 0 u2 0 u0 0 u0] */
          "psubw     (chroma_sub), %%mm1\n\t" /* u -= 128 */
          "psllw     $3, %%mm1\n\t" /* u << 3 */
          "psrld     $24, %%mm0\n\t" /* mm0 = [0 0 0 v2 0 0 0 v0] */
          "packssdw  %%mm0, %%mm0\n\t" /* mm0 = [0 v2 0 v0 0 v2 0 v0] */
          "punpcklwd %%mm0, %%mm0\n\t" /* mm0 = [0 v2 0 v2 0 v0 0 v0] */
          "psubw     (chroma_sub), %%mm0\n\t" /* v -= 128 */
          "psllw     $3, %%mm0\n\t" /* v << 3 */
          "movq      %%mm0, %%mm3\n\t"
          "pmulhw    (mmx_table), %%mm3\n\t" /* chroma_r */
          "paddw     %%mm2, %%mm3\n\t" /* mm3 = [r3 | r2 | r1 | r0] */
          "packuswb  %%mm3, %%mm3\n\t" /* mm3 = [r3 r2 r1 r0 r3 r2 r1 r0] */
          "movq      %%mm1, %%mm4\n\t"
          "pmulhw    (mmx_table + 24), %%mm4\n\t" /* chroma_b */
          "paddw     %%mm2, %%mm4\n\t" /* mm4 = [b3 | b2 | b1 | b0] */
          "packuswb  %%mm4, %%mm4\n\t" /* mm4 = [b3 b2 b1 b0 b3 b2 b1 b0] */
          "pmulhw    (mmx_table + 8), %%mm0\n\t"
          "pmulhw    (mmx_table + 16), %%mm1\n\t"
          "paddw     %%mm1, %%mm0\n\t" /* chroma_g */
          "psubw     %%mm0, %%mm2\n\t" /* mm2 = [g3 | g2 | g1 | g0] */
          "packuswb  %%mm2, %%mm2\n\t" /* mm2 = [g3 g2 g1 g0 g3 g2 g1 g0] */
          "punpcklbw %%mm7, %%mm3\n\t" /* mm3 = [0 r3 0 r2 0 r1 0 r0] */
          "punpcklbw %%mm2, %%mm4\n\t" /* mm4 = [g3 b3 g2 b2 g1 b1 g0 b0] */
          "movq      %%mm4, %%mm0\n\t" /* mm0 = [g3 b3 g2 b2 g1 b1 g0 b0] */
          "punpcklwd %%mm3, %%mm0\n\t" /* mm0 = [0 r1 g1 b1 0 r0 g0 b0] */
          "movq      %%mm0, (%0)\n\t" /* out first 2 pixels */
          "punpckhwd %%mm3, %%mm4\n\t" /* mm4 = [0 r3 g3 b3 0 r2 g2 b2] */
          "movq      %%mm4, 8(%0)\n\t" /* out second 2 pixels */
          "addl      $16, %0\n\t"
          "addl      $8, %1\n\t"
          "decl      %2\n\t"
          "jnz       1b\n\t"
          "emms"

          :: "r" (dst), "r" (src), "q" (n)
          : "memory");
}


static
DFB_PFUNCTION( yuy2, argb )
{
     uint8_t *src = frame->vo_frame.base[0];
     int      n   = (frame->width.cur * frame->height.cur) >> 2;

     __asm__ __volatile__(

          "pxor      %%mm7, %%mm7\n\t"
          ".align 16\n"
          "1:\n\t"
          "movq      (%1), %%mm0\n\t" /* mm1 = [v2 y3 u2 y2 v0 y1 u0 y0] */
          "movq      %%mm0, %%mm2\n\t" /* mm2 = [v2 y3 u2 y2 v0 y1 u0 y0] */
          "pand      (wmask), %%mm2\n\t" /* mm2 = [0 y3 0 y2 0 y1 0 y0] */
          "paddw     (mmx_table + 32), %%mm2\n\t" /* y + brightness */
          "psllw     $2, %%mm2\n\t" /* y << 2 */
          "pmulhw    (mmx_table + 40), %%mm2\n\t" /* y * contrast */
          "movq      %%mm0, %%mm1\n\t" /* mm1 = [v2 y3 u2 y2 v0 y1 u0 y0] */
          "psrld     $8, %%mm1\n\t" /* mm1 = [0 v2 y3 u2 0 v0 y1 u0] */
          "pand      (lmask), %%mm1\n\t" /* mm1 = [0 0 0 u2 0 0 0 u0] */
          "packssdw  %%mm1, %%mm1\n\t" /* mm1 = [0 u2 0 u0 0 u2 0 u0] */
          "punpcklwd %%mm1, %%mm1\n\t" /* mm1 = [0 u2 0 u2 0 u0 0 u0] */
          "psubw     (chroma_sub), %%mm1\n\t" /* u -= 128 */
          "psllw     $3, %%mm1\n\t" /* u << 3 */
          "psrld     $24, %%mm0\n\t" /* mm0 = [0 0 0 v2 0 0 0 v0] */
          "packssdw  %%mm0, %%mm0\n\t" /* mm0 = [0 v2 0 v0 0 v2 0 v0] */
          "punpcklwd %%mm0, %%mm0\n\t" /* mm0 = [0 v2 0 v2 0 v0 0 v0] */
          "psubw     (chroma_sub), %%mm0\n\t" /* v -= 128 */
          "psllw     $3, %%mm0\n\t" /* v << 3 */
          "movq      %%mm0, %%mm3\n\t"
          "pmulhw    (mmx_table), %%mm3\n\t" /* chroma_r */
          "paddw     %%mm2, %%mm3\n\t" /* mm3 = [r3 | r2 | r1 | r0] */
          "packuswb  %%mm3, %%mm3\n\t" /* mm3 = [r3 r2 r1 r0 r3 r2 r1 r0] */
          "movq      %%mm1, %%mm4\n\t"
          "pmulhw    (mmx_table + 24), %%mm4\n\t" /* chroma_b */
          "paddw     %%mm2, %%mm4\n\t" /* mm4 = [b3 | b2 | b1 | b0] */
          "packuswb  %%mm4, %%mm4\n\t" /* mm4 = [b3 b2 b1 b0 b3 b2 b1 b0] */
          "pmulhw    (mmx_table + 8), %%mm0\n\t"
          "pmulhw    (mmx_table + 16), %%mm1\n\t"
          "paddw     %%mm1, %%mm0\n\t" /* chroma_g */
          "psubw     %%mm0, %%mm2\n\t" /* mm2 = [g3 | g2 | g1 | g0] */
          "packuswb  %%mm2, %%mm2\n\t" /* mm2 = [g3 g2 g1 g0 g3 g2 g1 g0] */
          "punpcklbw %%mm7, %%mm3\n\t" /* mm3 = [0 r3 0 r2 0 r1 0 r0] */
          "por       (amask), %%mm3\n\t" /* mm3 = [ff r3 ff r2 ff r1 ff r0] */
          "punpcklbw %%mm2, %%mm4\n\t" /* mm4 = [g3 b3 g2 b2 g1 b1 g0 b0] */
          "movq      %%mm4, %%mm0\n\t" /* mm0 = [g3 b3 g2 b2 g1 b1 g0 b0] */
          "punpcklwd %%mm3, %%mm0\n\t" /* mm0 = [ff r1 g1 b1 ff r0 g0 b0] */
          "movq      %%mm0, (%0)\n\t" /* out first 2 pixels */
          "punpckhwd %%mm3, %%mm4\n\t" /* mm4 = [ff r3 g3 b3 ff r2 g2 b2] */
          "movq      %%mm4, 8(%0)\n\t" /* out second 2 pixels */
          "addl      $16, %0\n\t"
          "addl      $8, %1\n\t"
          "decl      %2\n\t"
          "jnz       1b\n\t"
          "emms"

          :: "r" (dst), "r" (src), "q" (n)
          : "memory");
}


/******************************** End of YV12 callbacks *********************************/


/******************************** Begin YV12 callbacks **********************************/

static
DFB_PFUNCTION( yv12, yuy2 )
{
     uint8_t *srcy  = frame->vo_frame.base[0];
     uint8_t *srcu  = frame->vo_frame.base[1];
     uint8_t *srcv  = frame->vo_frame.base[2];
     int      n     = frame->height.cur >> 1;
     int      line;

     __asm__ __volatile__(

          "movq      (mmx_table + 32), %%mm4\n\t" /* mm4 = brightness */
          "movq      (mmx_table + 40), %%mm5\n\t" /* mm5 = contrast */
          "movq      (mmx_table + 48), %%mm6\n\t" /* mm6 = saturation */
          "pxor      %%mm7, %%mm7\n\t"
          ".align 16\n"
          "1:\n\t"
          "movl      %4, %7\n\t"
          "shrl      $2, %7\n\t"
          ".align 16\n"
          "2:\n\t"
          "movd      (%1), %%mm0\n\t" /* mm0 = [0 0 0 0 y03 y02 y01 y00] */
          "punpcklbw %%mm7, %%mm0\n\t" /* mm0 = [0 y03 0 y02 0 y01 0 y00] */
          "movd      (%1, %4), %%mm1\n\t" /* mm1 = [0 0 0 0 y13 y12 y11 y10] */
          "punpcklbw %%mm7, %%mm1\n\t" /* mm1 = [0 y13 0 y12 0 y11 0 y10] */
          "testl     $3, %6\n\t"
          "jz        3f\n\t"
          "paddw     %%mm4, %%mm0\n\t" /* y + brightness */
          "psllw     $2, %%mm0\n\t" /* y << 2 */
          "pmulhw    %%mm5, %%mm0\n\t" /* y * contrast */
          "paddw     %%mm4, %%mm1\n\t" /* y + brightness */
          "psllw     $2, %%mm1\n\t" /* y << 2 */
          "pmulhw    %%mm5, %%mm1\n\t" /* y * contrast */
          ".align 8\n"
          "3:\n\t"
          "packuswb  %%mm1, %%mm0\n\t" /* mm0 = [y13 y12 y11 y10 y03 y02 y01 y00] */
          "movd      (%2), %%mm2\n\t" /* mm2 = [0 0 0 0 u3 u2 u1 u0] */
          "addl      $2, %2\n\t" /* srcu += 2 */
          "movd      (%3), %%mm3\n\t" /* mm3 = [0 0 0 0 v3 v2 v1 v0] */
          "addl      $2, %3\n\t" /* srcv += 2 */
          "testl     $4, %6\n\t"
          "jz        4f\n\t"
          "punpcklbw %%mm7, %%mm2\n\t" /* mm2 = [0 u3 0 u2 0 u1 0 u0] */
          "psubw     (chroma_sub), %%mm2\n\t" /* u - 128 */
          "psllw     $2, %%mm2\n\t" /* u << 2 */
          "pmulhw    %%mm6, %%mm2\n\t" /* u * saturation */
          "paddw     (chroma_sub), %%mm2\n\t" /* u + 128 */
          "packuswb  %%mm2, %%mm2\n\t" /* mm2 = [u3 u2 u1 u0 u3 u2 u1 u0] */
          "punpcklbw %%mm7, %%mm3\n\t" /* mm3 = [0 v3 0 v2 0 v1 0 v0] */
          "psubw     (chroma_sub), %%mm3\n\t" /* v - 128 */
          "psllw     $2, %%mm3\n\t" /* v << 2 */
          "pmulhw    %%mm6, %%mm3\n\t" /* v * saturation */
          "paddw     (chroma_sub), %%mm3\n\t" /* v + 128 */
          "packuswb  %%mm3, %%mm3\n\t" /* mm3 = [v3 v2 v1 v0 v3 v2 v1 v0] */
          ".align 8\n"
          "4:\n\t"
          "punpcklbw %%mm3, %%mm2\n\t" /* mm2 = [v3 u3 v2 u2 v1 u1 v0 u0] */
          "movq      %%mm0, %%mm1\n\t" /* mm1 = [y13 y12 y11 y10 y03 y02 y01 y00] */
          "punpcklbw %%mm2, %%mm1\n\t" /* mm1 = [v1 y03 u1 y02 v0 y01 u0 y00] */
          "movq      %%mm1, (%0)\n\t" /* out 2 pixels (line 1) */
          "punpckhdq %%mm0, %%mm0\n\t" /* mm0 = [y13 y12 y11 y10 y13 y12 y11 y10] */
          "punpcklbw %%mm2, %%mm0\n\t" /* mm0 = [v1 y13 u1 y12 v0 y11 u0 y10] */
          "movq      %%mm0, (%0, %5)\n\t" /* out 2 pixels (line 2) */
          "addl      $8, %0\n\t" /* dst += 8 */
          "addl      $4, %1\n\t" /* srcy += 4 */
          "decl      %7\n\t"
          "jnz       2b\n\t"
          "addl      %5, %0\n\t"
          "addl      %4, %1\n\t"
          "decl      %8\n\t"
          "jnz       1b\n\t"
          "emms"

          :: "r" (dst), "r" (srcy), "r" (srcu), "r" (srcv),
             "r" (frame->width.cur), "r" (pitch),
             "r" (this->mixer.modified), "m" (line), "m" (n)
          : "memory");
}


static
DFB_PFUNCTION( yv12, uyvy )
{
     uint8_t *srcy  = frame->vo_frame.base[0];
     uint8_t *srcu  = frame->vo_frame.base[1];
     uint8_t *srcv  = frame->vo_frame.base[2];
     int      n     = frame->height.cur >> 1;
     int      line;

     __asm__ __volatile__(

          "movq      (mmx_table + 32), %%mm4\n\t" /* mm4 = brightness */
          "movq      (mmx_table + 40), %%mm5\n\t" /* mm5 = contrast */
          "movq      (mmx_table + 48), %%mm6\n\t" /* mm6 = saturation */
          "pxor      %%mm7, %%mm7\n\t"
          ".align 16\n"
          "1:\n\t"
          "movl      %4, %7\n\t"
          "shrl      $2, %7\n\t"
          ".align 16\n"
          "2:\n\t"
          "movd      (%1), %%mm0\n\t" /* mm0 = [0 0 0 0 y03 y02 y01 y00] */
          "punpcklbw %%mm7, %%mm0\n\t" /* mm0 = [0 y03 0 y02 0 y01 0 y00] */
          "movd      (%1, %4), %%mm1\n\t" /* mm1 = [0 0 0 0 y13 y12 y10 y10] */
          "punpcklbw %%mm7, %%mm1\n\t" /* mm1 = [0 y13 0 y12 0 y11 0 y10] */
          "testl     $3, %6\n\t"
          "jz        3f\n\t"
          "paddw     %%mm4, %%mm0\n\t" /* y + brightness */
          "psllw     $2, %%mm0\n\t" /* y << 2 */
          "pmulhw    %%mm5, %%mm0\n\t" /* y * contrast */
          "paddw     %%mm4, %%mm1\n\t" /* y + brightness */
          "psllw     $2, %%mm1\n\t" /* y << 2 */
          "pmulhw    %%mm5, %%mm1\n\t" /* y * contrast */
          ".align 8\n"
          "3:\n\t"
          "packuswb  %%mm1, %%mm0\n\t" /* mm0 = [y13 y12 y10 y10 y03 y02 y01 y00] */
          "movd      (%2), %%mm2\n\t" /* mm2 = [0 0 0 0 u3 u2 u1 u0] */
          "addl      $2, %2\n\t" /* srcu += 2 */
          "movd      (%3), %%mm3\n\t" /* mm3 = [0 0 0 0 v3 v2 v1 v0] */
          "addl      $2, %3\n\t" /* srcv += 2 */
          "testl     $4, %6\n\t"
          "jz        4f\n\t"
          "punpcklbw %%mm7, %%mm2\n\t" /* mm2 = [0 u3 0 u2 0 u1 0 u0] */
          "psubw     (chroma_sub), %%mm2\n\t" /* u - 128 */
          "psllw     $2, %%mm2\n\t" /* u << 2 */
          "pmulhw    %%mm6, %%mm2\n\t" /* u * saturation */
          "paddw     (chroma_sub), %%mm2\n\t" /* u + 128 */
          "packuswb  %%mm2, %%mm2\n\t" /* mm2 = [u3 u2 u1 u0 u3 u2 u1 u0] */
          "punpcklbw %%mm7, %%mm3\n\t" /* mm3 = [0 v3 0 v2 0 v1 0 v0] */
          "psubw     (chroma_sub), %%mm3\n\t" /* v - 128 */
          "psllw     $2, %%mm3\n\t" /* v << 2 */
          "pmulhw    %%mm6, %%mm3\n\t" /* v * saturation */
          "paddw     (chroma_sub), %%mm3\n\t" /* v + 128 */
          "packuswb  %%mm3, %%mm3\n\t" /* mm3 = [v3 v2 v1 v0 v3 v2 v1 v0] */
          ".align 8\n"
          "4:\n\t"
          "punpcklbw %%mm3, %%mm2\n\t" /* mm2 = [v3 u3 v2 u2 v1 u1 v0 u0] */
          "movq      %%mm2, %%mm3\n\t" /* mm3 = [v3 u3 v2 u2 v1 u1 v0 u0] */
          "punpcklbw %%mm0, %%mm3\n\t" /* mm3 = [y03 v1 y02 u1 y01 v0 y00 u0] */
          "movq      %%mm3, (%0)\n\t"
          "punpckhdq %%mm0, %%mm0\n\t" /* mm0 = [y13 y12 y10 y10 y13 y12 y10 y10] */
          "punpcklbw %%mm0, %%mm2\n\t" /* mm2 = [y13 v1 y12 u1 y11 v0 y10 u0] */
          "movq      %%mm2, (%0, %5)\n\t"
          "addl      $8, %0\n\t" /* dst += 8 */
          "addl      $4, %1\n\t" /* srcy += 4 */
          "decl      %7\n\t"
          "jnz       2b\n\t"
          "addl      %5, %0\n\t"
          "addl      %4, %1\n\t"
          "decl      %8\n\t"
          "jnz       1b\n\t"
          "emms"

          :: "r" (dst), "r" (srcy), "r" (srcu), "r" (srcv),
             "r" (frame->width.cur), "r" (pitch),
             "r" (this->mixer.modified), "m" (line), "m" (n)
          : "memory");
}


static
DFB_PFUNCTION( yv12, yv12 )
{
     uint8_t *srcy  = frame->vo_frame.base[0];
     uint8_t *srcu;
     uint8_t *srcv;
     uint8_t *dsty  = dst;
     uint8_t *dstu;
     uint8_t *dstv;
     int      ys    = pitch * frame->height.cur;
     int      crs   = (pitch * frame->height.cur) >> 2;
     int      mmf   = this->mixer.modified;

     if (frame->dstfmt.cur == DSPF_YV12)
     {
          srcu = frame->vo_frame.base[1];
          srcv = frame->vo_frame.base[2];
          dstv  = (uint8_t*) dsty + ys;
          dstu  = (uint8_t*) dstv + crs;
     } else
     {
          srcu = frame->vo_frame.base[2];
          srcv = frame->vo_frame.base[1];
          dstu  = (uint8_t*) dsty + ys;
          dstv  = (uint8_t*) dstu + crs;
     }

     if (mmf & (MMF_B | MMF_C))
     {
          __asm__ __volatile__(

               "movq      (mmx_table + 32), %%mm1\n\t" /* mm1 = brightness */
               "movq      (mmx_table + 40), %%mm2\n\t" /* mm2 = contrast */
               "pxor      %%mm7, %%mm7\n\t"
               ".align 16\n"
               "1:\n\t"
               "movd      (%1), %%mm0\n\t" /* mm0 = [0 0 0 0 y3 y2 y1 y0] */
               "punpcklbw %%mm7, %%mm0\n\t" /* mm0 = [0 y3 0 y2 0 y1 0 y0] */
               "paddw     %%mm1, %%mm0\n\t" /* y + brightness */
               "psllw     $2, %%mm0\n\t" /* y << 2 */
               "pmulhw    %%mm2, %%mm0\n\t" /* y * contrast */
               "packuswb  %%mm0, %%mm0\n\t" /* mm0 = [y3 y2 y1 y0 y3 y2 y1 y0] */
               "movd      %%mm0, (%0)\n\t"
               "addl      $4, %0\n\t"
               "addl      $4, %1\n\t"
               "decl      %2\n\t"
               "jnz       1b\n\t"
               "emms"

               :: "r" (dsty), "r" (srcy),
                  "q" (ys >> 2)
               :  "memory");
     } else
          xine_fast_memcpy( dsty, srcy, ys );

     if (mmf & MMF_S)
     {
          __asm__ __volatile__(

               "movq      (chroma_sub), %%mm2\n\t" /* mm2 = [0 128 0 128...] */
               "movq      (mmx_table + 48), %%mm3\n\t" /* mm3 = saturation */
               "pxor      %%mm7, %%mm7\n\t"
               ".align 16\n"
               "1:\n\t"
               "movd      (%2), %%mm0\n\t" /* mm0 = [0 0 0 0 u3 u2 u1 u0] */
               "punpcklbw %%mm7, %%mm0\n\t" /* mm0 = [0 u3 0 u2 0 u1 0 u0] */
               "psubw     %%mm2, %%mm0\n\t" /* u - 128 */
               "psllw     $2, %%mm0\n\t" /* u << 2 */
               "pmulhw    %%mm3, %%mm0\n\t" /* u * saturation */
               "paddw     %%mm2, %%mm0\n\t" /* u + 128 */
               "movd      (%3), %%mm1\n\t" /* mm1 = [0 0 0 0 v3 v2 v1 v0] */
               "punpcklbw %%mm7, %%mm1\n\t" /* mm1 = [0 v3 0 v2 0 v1 0 v0] */
               "psubw     %%mm2, %%mm1\n\t" /* v - 128 */
               "psllw     $2, %%mm1\n\t" /* v << 2 */
               "pmulhw    %%mm3, %%mm1\n\t" /* v * saturation */
               "paddw     %%mm2, %%mm1\n\t" /* v + 128 */
               "packuswb  %%mm1, %%mm0\n\t" /* mm0 = [v3 v2 v1 v0 u3 u2 u1 u0] */
               "movd      %%mm0, (%0)\n\t"
               "punpckhdq %%mm0, %%mm0\n\t" /* mm0 = [v3 v2 v1 v0 v3 v2 v1 v0] */
               "movd      %%mm0, (%1)\n\t"
               "addl      $4, %0\n\t"
               "addl      $4, %1\n\t"
               "addl      $4, %2\n\t"
               "addl      $4, %3\n\t"
               "decl      %4\n\t"
               "jnz       1b\n\t"
               "emms"

               :: "r" (dstu), "r" (dstv), "r" (srcu), "r" (srcv),
                  "q" (crs >> 2)
               :  "memory");
     } else
     {
          xine_fast_memcpy( dstu, srcu, crs );
          xine_fast_memcpy( dstv, srcv, crs );
     }
}


static
DFB_PFUNCTION( yv12, rgb8 )
{
     uint8_t *srcy = frame->vo_frame.base[0];
     uint8_t *srcu = frame->vo_frame.base[1];
     uint8_t *srcv = frame->vo_frame.base[2];
     int      n    = (frame->width.cur * frame->height.cur) >> 3;
     int      line = frame->width.cur >> 2;

     __asm__ __volatile__(

          "pxor      %%mm7, %%mm7\n\t"
          ".align 16\n"
          "1:\n\t"
          "movd      (%1), %%mm3\n\t" /* mm3 = [0 0 0 0 y03 y02 y01 y00] */
          "punpcklbw %%mm7, %%mm3\n\t" /* mm3 = [0 y03 0 y02 0 y01 0 y00] */
          "paddw     (mmx_table + 32), %%mm3\n\t" /* y + brightness */
          "psllw     $2, %%mm3\n\t" /* y << 2 */
          "pmulhw    (mmx_table + 40), %%mm3\n\t" /* y * contrast */
          "movd      (%1, %4), %%mm2\n\t" /* mm2 = [0 0 0 0 y13 y12 y11 y10] */
          "punpcklbw %%mm7, %%mm2\n\t" /* mm2 = [0 y13 0 y12 0 y11 0 y10] */
          "paddw     (mmx_table + 32), %%mm2\n\t" /* y + brightness */
          "psllw     $2, %%mm2\n\t" /* y << 2 */
          "pmulhw    (mmx_table + 40), %%mm2\n\t" /* y * contrast */
          "movd      (%2), %%mm1\n\t" /* mm1 = [0 0 0 0 u3 u2 u1 u0] */
          "punpcklbw %%mm1, %%mm1\n\t" /* mm1 = [u3 u3 u2 u2 u1 u1 u0 u0] */
          "punpcklbw %%mm7, %%mm1\n\t" /* mm1 = [0 u1 0 u1 0 u0 0 u0] */
          "psubw     (chroma_sub), %%mm1\n\t" /* u -= 128 */
          "psllw     $3, %%mm1\n\t" /* u << 3 */
          "movd      (%3), %%mm0\n\t" /* mm0 = [0 0 0 0 v3 v2 v1 v0] */
          "punpcklbw %%mm0, %%mm0\n\t" /* mm0 = [v3 v3 v2 v2 v1 v1 v0 v0] */
          "punpcklbw %%mm7, %%mm0\n\t" /* mm0 = [0 v1 0 v1 0 v0 0 v0] */
          "psubw     (chroma_sub), %%mm0\n\t" /* v -= 128 */
          "psllw     $3, %%mm0\n\t" /* v << 3 */
          "movq      %%mm0, %%mm4\n\t" /* mm4 = [0 v1 0 v1 0 v0 0 v0] */
          "pmulhw    (mmx_table), %%mm4\n\t" /* chroma_r */
          "movq      %%mm4, %%mm5\n\t"
          "paddw     %%mm3, %%mm4\n\t" /* mm4 = [r03 | r02 | r01 | r00] */
          "paddw     %%mm2, %%mm5\n\t" /* mm5 = [r13 | r12 | r11 | r10] */
          "packuswb  %%mm5, %%mm4\n\t" /* mm4 = [r13 r12 r11 r10 r03 r02 r01 r00] */
          "pand      (b3_mask), %%mm4\n\t" /* red 8 */
          "movq      %%mm1, %%mm5\n\t" /* mm5 = [0 u1 0 u1 0 u0 0 u0] */
          "pmulhw    (mmx_table + 24), %%mm5\n\t" /* chroma_b */
          "movq      %%mm5, %%mm6\n\t"
          "paddw     %%mm3, %%mm5\n\t" /* mm5 = [b03 | b02 | b01 | b00] */
          "paddw     %%mm2, %%mm6\n\t" /* mm6 = [b13 | b12 | b11 | b00] */
          "packuswb  %%mm6, %%mm5\n\t" /* mm5 = [b13 b12 b11 b10 b03 b02 b01 b00] */
          "pmulhw    (mmx_table + 8), %%mm0\n\t"
          "pmulhw    (mmx_table + 16), %%mm1\n\t"
          "paddw     %%mm1, %%mm0\n\t" /* chroma_g */
          "psubw     %%mm0, %%mm3\n\t" /* mm3 = [g03 | g02 | g01 | g00] */
          "psubw     %%mm0, %%mm2\n\t" /* mm2 = [g13 | g12 | g11 | g10] */
          "packuswb  %%mm2, %%mm3\n\t" /* mm3 = [g13 g12 g11 g10 g03 g02 g01 g00] */
          "pand      (b3_mask), %%mm3\n\t" /* green 8 */
          "movq      %%mm4, %%mm0\n\t" /* mm0 = [r13 r12 r11 r10 r03 r02 r01 r00] */
          "punpcklbw %%mm7, %%mm0\n\t" /* mm0 = [0 r03 0 r02 0 r01 0 r00] */
          "movq      %%mm5, %%mm1\n\t" /* mm1 = [b13 b12 b11 b10 b03 b02 b01 b00] */
          "punpcklbw %%mm7, %%mm1\n\t" /* mm1 = [0 b03 0 b02 0 b01 0 b00] */
          "psrlw     $6, %%mm1\n\t" /* blue 8 */
          "por       %%mm1, %%mm0\n\t"
          "movq      %%mm3, %%mm1\n\t" /* mm1 = [g13 g12 g11 g10 g03 g02 g01 g00] */
          "punpcklbw %%mm7, %%mm1\n\t" /* mm1 = [0 g03 0 g02 0 g01 0 g00] */
          "psrlw     $3, %%mm1\n\t"
          "por       %%mm1, %%mm0\n\t"
          "packuswb  %%mm0, %%mm0\n\t" /* rgb8 */
          "movd      %%mm0, (%0)\n\t" /* out 4 pixels (line 1) */
          "punpckhbw %%mm7, %%mm4\n\t" /* mm4 = [0 r13 0 r12 0 r11 0 r10] */
          "punpckhbw %%mm7, %%mm5\n\t" /* mm5 = [0 b13 0 b12 0 b11 0 b10] */
          "psrlw     $6, %%mm5\n\t" /* blue 8 */
          "por       %%mm5, %%mm4\n\t"
          "punpckhbw %%mm7, %%mm3\n\t" /* mm3 = [0 g13 0 g12 0 g11 0 g10] */
          "psrlw     $3, %%mm3\n\t"
          "por       %%mm3, %%mm4\n\t"
          "packuswb  %%mm4, %%mm4\n\t" /* rgb 8 */
          "movd      %%mm4, (%0, %4)\n\t" /* out 4 pixels (line 2) */
          "addl      $4, %0\n\t"
          "addl      $4, %1\n\t"
          "addl      $2, %2\n\t"
          "addl      $2, %3\n\t"
          "decl      %5\n\t" /* --line */
          "jnz       2f\n\t"
          "movl      %4, %5\n\t" /* if (!line) { */
          "shrl      $2, %5\n\t" /* line = frame->width.cur / 4 */
          "addl      %4, %0\n\t" /* dst += pitch */
          "addl      %4, %1\n\t" /* srcy += frame->width.cur } */
          "2:\n\t"
          "decl      %6\n\t"
          "jnz       1b\n\t"
          "emms"

          :: "r" (dst), "r" (srcy), "r" (srcu), "r" (srcv),
             "q" (frame->width.cur), "q" (line), "m" (n)
          : "memory");
}

     
static
DFB_PFUNCTION( yv12, rgb15 )
{
     uint8_t *srcy = frame->vo_frame.base[0];
     uint8_t *srcu = frame->vo_frame.base[1];
     uint8_t *srcv = frame->vo_frame.base[2];
     int      n    = (frame->width.cur * frame->height.cur) >> 3;
     int      line = frame->width.cur >> 2;

     __asm__ __volatile__(

          "pxor      %%mm7, %%mm7\n\t"
          ".align 16\n"
          "1:\n\t"
          "movd      (%1), %%mm3\n\t" /* mm3 = [0 0 0 0 y03 y02 y01 y00] */
          "punpcklbw %%mm7, %%mm3\n\t" /* mm3 = [0 y03 0 y02 0 y01 0 y00] */
          "paddw     (mmx_table + 32), %%mm3\n\t" /* y + brightness */
          "psllw     $2, %%mm3\n\t" /* y << 2 */
          "pmulhw    (mmx_table + 40), %%mm3\n\t" /* y * contrast */
          "movd      (%1, %4), %%mm2\n\t" /* mm2 = [0 0 0 0 y13 y12 y11 y10] */
          "punpcklbw %%mm7, %%mm2\n\t" /* mm2 = [0 y13 0 y12 0 y11 0 y10] */
          "paddw     (mmx_table + 32), %%mm2\n\t" /* y + brightness */
          "psllw     $2, %%mm2\n\t" /* y << 2 */
          "pmulhw    (mmx_table + 40), %%mm2\n\t" /* y * contrast */
          "movd      (%2), %%mm1\n\t" /* mm1 = [0 0 0 0 u3 u2 u1 u0] */
          "punpcklbw %%mm1, %%mm1\n\t" /* mm1 = [u3 u3 u2 u2 u1 u1 u0 u0] */
          "punpcklbw %%mm7, %%mm1\n\t" /* mm1 = [0 u1 0 u1 0 u0 0 u0] */
          "psubw     (chroma_sub), %%mm1\n\t" /* u -= 128 */
          "psllw      $3, %%mm1\n\t" /* u << 3 */
          "movd      (%3), %%mm0\n\t" /* mm0 = [0 0 0 0 v3 v2 v1 v0] */
          "punpcklbw %%mm0, %%mm0\n\t" /* mm0 = [v3 v3 v2 v2 v1 v1 v0 v0] */
          "punpcklbw %%mm7, %%mm0\n\t" /* mm0 = [0 v1 0 v1 0 v0 0 v0] */
          "psubw     (chroma_sub), %%mm0\n\t" /* v -= 128 */
          "psllw     $3, %%mm0\n\t" /* v << 3 */
          "movq      %%mm0, %%mm4\n\t" /* mm4 = [0 v1 0 v1 0 v0 0 v0] */
          "pmulhw    (mmx_table), %%mm4\n\t" /* chroma_r*/
          "movq      %%mm4, %%mm5\n\t"
          "paddw     %%mm3, %%mm4\n\t" /* mm4 = [r03 | r02 | r01 | r00] */
          "paddw     %%mm2, %%mm5\n\t" /* mm5 = [r13 | r12 | r11 | r10] */
          "packuswb  %%mm5, %%mm4\n\t" /* mm4 = [r13 r12 r11 r10 r03 r02 r01 r00] */
          "pand      (b5_mask), %%mm4\n\t" /* red 15 */
          "movq      %%mm1, %%mm5\n\t" /* mm5 = [0 u1 0 u1 0 u0 0 u0] */
          "pmulhw    (mmx_table + 24), %%mm5\n\t" /* chroma_b */
          "movq      %%mm5, %%mm6\n\t"
          "paddw     %%mm3, %%mm5\n\t" /* mm5 = [b03 | b02 | b01 | b00] */
          "paddw     %%mm2, %%mm6\n\t" /* mm6 = [b13 | b12 | b11 | b00] */
          "packuswb  %%mm6, %%mm5\n\t" /* mm5 = [b13 b12 b11 b10 b03 b02 b01 b00] */
          "pmulhw    (mmx_table + 8), %%mm0\n\t"
          "pmulhw    (mmx_table + 16), %%mm1\n\t"
          "paddw     %%mm1, %%mm0\n\t" /* chroma_g */
          "psubw     %%mm0, %%mm3\n\t" /* mm3 = [g03 | g02 | g01 | g00] */
          "psubw     %%mm0, %%mm2\n\t" /* mm2 = [g13 | g12 | g11 | g10] */
          "packuswb  %%mm2, %%mm3\n\t" /* mm3 = [g13 g12 g11 g10 g03 g02 g01 g00] */
          "pand      (b5_mask), %%mm3\n\t" /* green 15 */
          "movq      %%mm4, %%mm0\n\t" /* mm0 = [r13 r12 r11 r10 r03 r02 r01 r00] */
          "punpcklbw %%mm7, %%mm0\n\t" /* mm0 = [0 r03 0 r02 0 r01 0 r00] */
          "psllw     $7, %%mm0\n\t"
          "movq      %%mm5, %%mm1\n\t" /* mm1 = [b13 b12 b11 b10 b03 b02 b01 b00] */
          "punpcklbw %%mm7, %%mm1\n\t" /* mm1 = [0 b03 0 b02 0 b01 0 b00] */
          "psrlw     $3, %%mm1\n\t" /* blue 15 */
          "por       %%mm1, %%mm0\n\t"
          "movq      %%mm3, %%mm1\n\t" /* mm1 = [g13 g12 g11 g10 g03 g02 g01 g00] */
          "punpcklbw %%mm7, %%mm1\n\t" /* mm1 = [0 g03 0 g02 0 g01 0 g00] */
          "psllw     $2, %%mm1\n\t"
          "por       %%mm1, %%mm0\n\t" /* rgb15 */
          "movq      %%mm0, (%0)\n\t" /* out 4 pixels (line 1) */
          "punpckhbw %%mm7, %%mm4\n\t" /* mm4 = [0 r13 0 r12 0 r11 0 r10] */
          "psllw     $7, %%mm4\n\t"
          "punpckhbw %%mm7, %%mm5\n\t" /* mm5 = [0 b13 0 b12 0 b11 0 b10] */
          "psrlw     $3, %%mm5\n\t" /* blue 15 */
          "por       %%mm5, %%mm4\n\t"
          "punpckhbw %%mm7, %%mm3\n\t" /* mm3 = [0 g13 0 g12 0 g11 0 g10] */
          "psllw     $2, %%mm3\n\t"
          "por       %%mm3, %%mm4\n\t" /* rgb15 */
          "movq      %%mm4, (%0, %5)\n\t" /* out 4 pixels (line 2) */
          "addl      $8, %0\n\t"
          "addl      $4, %1\n\t"
          "addl      $2, %2\n\t"
          "addl      $2, %3\n\t"
          "decl      %6\n\t" /* --line */
          "jnz       2f\n\t"
          "movl      %4, %6\n\t" /* if (!line) { */
          "shrl      $2, %6\n\t" /* line = frame->width.cur / 4 */
          "addl      %5, %0\n\t" /* dst += pitch */
          "addl      %4, %1\n\t" /* srcy += frame->width.cur } */
          "2:\n\t"
          "decl      %7\n\t"
          "jnz       1b\n\t"
          "emms"

          :: "r" (dst), "r" (srcy), "r" (srcu), "r" (srcv),
             "q" (frame->width.cur), "q" (pitch), "m" (line), "m" (n)
          : "memory");
}


static
DFB_PFUNCTION( yv12, rgb16 )
{
     uint8_t *srcy = frame->vo_frame.base[0];
     uint8_t *srcu = frame->vo_frame.base[1];
     uint8_t *srcv = frame->vo_frame.base[2];
     int      n    = (frame->width.cur * frame->height.cur) >> 3;
     int      line = frame->width.cur >> 2;

     __asm__ __volatile__(

          "pxor      %%mm7, %%mm7\n\t"
          ".align 16\n"
          "1:\n\t"
          "movd      (%1), %%mm3\n\t" /* mm3 = [0 0 0 0 y03 y02 y01 y00] */
          "punpcklbw %%mm7, %%mm3\n\t" /* mm3 = [0 y03 0 y02 0 y01 0 y00] */
          "paddw     (mmx_table + 32), %%mm3\n\t" /* y + brightness */
          "psllw     $2, %%mm3\n\t" /* y << 2 */
          "pmulhw    (mmx_table + 40), %%mm3\n\t" /* y * contrast */
          "movd      (%1, %4), %%mm2\n\t" /* mm2 = [0 0 0 0 y13 y12 y11 y10] */
          "punpcklbw %%mm7, %%mm2\n\t" /* mm2 = [0 y13 0 y12 0 y11 0 y10] */
          "paddw     (mmx_table + 32), %%mm2\n\t" /* y + brightness */
          "psllw     $2, %%mm2\n\t" /* y << 2 */
          "pmulhw    (mmx_table + 40), %%mm2\n\t" /* y * contrast */
          "movd      (%2), %%mm1\n\t" /* mm1 = [0 0 0 0 u3 u2 u1 u0] */
          "punpcklbw %%mm1, %%mm1\n\t" /* mm1 = [u3 u3 u2 u2 u1 u1 u0 u0] */
          "punpcklbw %%mm7, %%mm1\n\t" /* mm1 = [0 u1 0 u1 0 u0 0 u0] */
          "psubw     (chroma_sub), %%mm1\n\t" /* u -= 128 */
          "psllw      $3, %%mm1\n\t" /* u << 3 */
          "movd      (%3), %%mm0\n\t" /* mm0 = [0 0 0 0 v3 v2 v1 v0] */
          "punpcklbw %%mm0, %%mm0\n\t" /* mm0 = [v3 v3 v2 v2 v1 v1 v0 v0] */
          "punpcklbw %%mm7, %%mm0\n\t" /* mm0 = [0 v1 0 v1 0 v0 0 v0] */
          "psubw     (chroma_sub), %%mm0\n\t" /* v -= 128 */
          "psllw     $3, %%mm0\n\t" /* v << 3 */
          "movq      %%mm0, %%mm4\n\t" /* mm4 = [0 v1 0 v1 0 v0 0 v0] */
          "pmulhw    (mmx_table), %%mm4\n\t" /* chroma_r*/
          "movq      %%mm4, %%mm5\n\t"
          "paddw     %%mm3, %%mm4\n\t" /* mm4 = [r03 | r02 | r01 | r00] */
          "paddw     %%mm2, %%mm5\n\t" /* mm5 = [r13 | r12 | r11 | r10] */
          "packuswb  %%mm5, %%mm4\n\t" /* mm4 = [r13 r12 r11 r10 r03 r02 r01 r00] */
          "pand      (b5_mask), %%mm4\n\t" /* red 16 */
          "movq      %%mm1, %%mm5\n\t" /* mm5 = [0 u1 0 u1 0 u0 0 u0] */
          "pmulhw    (mmx_table + 24), %%mm5\n\t" /* chroma_b */
          "movq      %%mm5, %%mm6\n\t"
          "paddw     %%mm3, %%mm5\n\t" /* mm5 = [b03 | b02 | b01 | b00] */
          "paddw     %%mm2, %%mm6\n\t" /* mm6 = [b13 | b12 | b11 | b00] */
          "packuswb  %%mm6, %%mm5\n\t" /* mm5 = [b13 b12 b11 b10 b03 b02 b01 b00] */
          "pmulhw    (mmx_table + 8), %%mm0\n\t"
          "pmulhw    (mmx_table + 16), %%mm1\n\t"
          "paddw     %%mm1, %%mm0\n\t" /* chroma_g */
          "psubw     %%mm0, %%mm3\n\t" /* mm3 = [g03 | g02 | g01 | g00] */
          "psubw     %%mm0, %%mm2\n\t" /* mm2 = [g13 | g12 | g11 | g10] */
          "packuswb  %%mm2, %%mm3\n\t" /* mm3 = [g13 g12 g11 g10 g03 g02 g01 g00] */
          "pand      (b6_mask), %%mm3\n\t" /* green 16 */
          "movq      %%mm4, %%mm0\n\t" /* mm0 = [r13 r12 r11 r10 r03 r02 r01 r00] */
          "punpcklbw %%mm7, %%mm0\n\t" /* mm0 = [0 r03 0 r02 0 r01 0 r00] */
          "psllw     $8, %%mm0\n\t"
          "movq      %%mm5, %%mm1\n\t" /* mm1 = [b13 b12 b11 b10 b03 b02 b01 b00] */
          "punpcklbw %%mm7, %%mm1\n\t" /* mm1 = [0 b03 0 b02 0 b01 0 b00] */
          "psrlw     $3, %%mm1\n\t" /* blue 16 */
          "por       %%mm1, %%mm0\n\t"
          "movq      %%mm3, %%mm1\n\t" /* mm1 = [g13 g12 g11 g10 g03 g02 g01 g00] */
          "punpcklbw %%mm7, %%mm1\n\t" /* mm1 = [0 g03 0 g02 0 g01 0 g00] */
          "psllw     $3, %%mm1\n\t"
          "por       %%mm1, %%mm0\n\t" /* rgb16 */
          "movq      %%mm0, (%0)\n\t" /* out 4 pixels (line 1) */
          "punpckhbw %%mm7, %%mm4\n\t" /* mm4 = [0 r13 0 r12 0 r11 0 r10] */
          "psllw     $8, %%mm4\n\t"
          "punpckhbw %%mm7, %%mm5\n\t" /* mm5 = [0 b13 0 b12 0 b11 0 b10] */
          "psrlw     $3, %%mm5\n\t" /* blue 16 */
          "por       %%mm5, %%mm4\n\t"
          "punpckhbw %%mm7, %%mm3\n\t" /* mm3 = [0 g13 0 g12 0 g11 0 g10] */
          "psllw     $3, %%mm3\n\t"
          "por       %%mm3, %%mm4\n\t" /* rgb16 */
          "movq      %%mm4, (%0, %5)\n\t" /* out 4 pixels (line 2) */
          "addl      $8, %0\n\t"
          "addl      $4, %1\n\t"
          "addl      $2, %2\n\t"
          "addl      $2, %3\n\t"
          "decl      %6\n\t" /* --line */
          "jnz       2f\n\t"
          "movl      %4, %6\n\t" /* if (!line) { */
          "shrl      $2, %6\n\t" /* line = frame->width.cur / 4 */
          "addl      %5, %0\n\t" /* dst += pitch */
          "addl      %4, %1\n\t" /* srcy += frame->width.cur } */
          "2:\n\t"
          "decl      %7\n\t"
          "jnz       1b\n\t"
          "emms"

          :: "r" (dst), "r" (srcy), "r" (srcu), "r" (srcv),
             "q" (frame->width.cur), "q" (pitch), "m" (line), "m" (n)
          : "memory");
}


static
DFB_PFUNCTION( yv12, rgb24 )
{
     uint8_t *srcy = frame->vo_frame.base[0];
     uint8_t *srcu = frame->vo_frame.base[1];
     uint8_t *srcv = frame->vo_frame.base[2];
     int      n    = (frame->width.cur * frame->height.cur) >> 3;
     int      line = frame->width.cur >> 2;

     __asm__ __volatile__(

          "pxor      %%mm7, %%mm7\n\t"
          ".align 16\n"
          "1:\n\t"
          "movd      (%1), %%mm3\n\t" /* mm3 = [0 0 0 0 y03 y02 y01 y00] */
          "punpcklbw %%mm7, %%mm3\n\t" /* mm3 = [0 y03 0 y02 0 y01 0 y00] */
          "paddw     (mmx_table + 32), %%mm3\n\t" /* y + brightness */
          "psllw     $2, %%mm3\n\t" /* y << 2 */
          "pmulhw    (mmx_table + 40), %%mm3\n\t" /* y * contrast */
          "movd      (%1, %4), %%mm2\n\t" /* mm2 = [0 0 0 0 y13 y12 y11 y10] */
          "punpcklbw %%mm7, %%mm2\n\t" /* mm2 = [0 y13 0 y12 0 y11 0 y10] */
          "paddw     (mmx_table + 32), %%mm2\n\t" /* y + brightness */
          "psllw     $2, %%mm2\n\t" /* y << 2 */
          "pmulhw    (mmx_table + 40), %%mm2\n\t" /* y * contrast */
          "movd      (%2), %%mm1\n\t" /* mm1 = [0 0 0 0 u3 u2 u1 u0] */
          "punpcklbw %%mm1, %%mm1\n\t" /* mm1 = [u3 u3 u2 u2 u1 u1 u0 u0] */
          "punpcklbw %%mm7, %%mm1\n\t" /* mm1 = [0 u1 0 u1 0 u0 0 u0] */
          "psubw     (chroma_sub), %%mm1\n\t" /* u -= 128 */
          "psllw     $3, %%mm1\n\t" /* u << 3 */
          "movd      (%3), %%mm0\n\t" /* mm0 = [0 0 0 0 v3 v2 v1 v0] */
          "punpcklbw %%mm0, %%mm0\n\t" /* mm0 = [v3 v3 v2 v2 v1 v1 v0 v0] */
          "punpcklbw %%mm7, %%mm0\n\t" /* mm0 = [0 v1 0 v1 0 v0 0 v0] */
          "psubw     (chroma_sub), %%mm0\n\t" /* v -= 128 */
          "psllw     $3, %%mm0\n\t" /* v << 3 */
          "movq      %%mm0, %%mm4\n\t" /* mm4 = [0 v1 0 v1 0 v0 0 v0] */
          "pmulhw    (mmx_table), %%mm4\n\t" /* chroma_r*/
          "movq      %%mm4, %%mm5\n\t"
          "paddw     %%mm3, %%mm4\n\t" /* mm4 = [r03 | r02 | r01 | r00] */
          "paddw     %%mm2, %%mm5\n\t" /* mm5 = [r13 | r12 | r11 | r10] */
          "packuswb  %%mm5, %%mm4\n\t" /* mm4 = [r13 r12 r11 r10 r03 r02 r01 r00] */
          "movq      %%mm1, %%mm5\n\t" /* mm5 = [0 u1 0 u1 0 u0 0 u0] */
          "pmulhw    (mmx_table + 24), %%mm5\n\t" /* chroma_b */
          "movq      %%mm5, %%mm6\n\t"
          "paddw     %%mm3, %%mm5\n\t" /* mm5 = [b03 | b02 | b01 | b00] */
          "paddw     %%mm2, %%mm6\n\t" /* mm6 = [b13 | b12 | b11 | b00] */
          "packuswb  %%mm6, %%mm5\n\t" /* mm5 = [b13 b12 b11 b10 b03 b02 b01 b00] */
          "pmulhw    (mmx_table + 8), %%mm0\n\t"
          "pmulhw    (mmx_table + 16), %%mm1\n\t"
          "paddw     %%mm1, %%mm0\n\t" /* chorma_g */
          "psubw     %%mm0, %%mm3\n\t" /* mm3 = [g03 | g02 | g01 | g00] */
          "psubw     %%mm0, %%mm2\n\t" /* mm2 = [g13 | g12 | g11 | g10] */
          "packuswb  %%mm2, %%mm3\n\t" /* mm3 = [g13 g12 g11 g10 g03 g02 g01 g00] */
          "movq      %%mm5, %%mm0\n\t" /* mm0 = [b13 b12 b11 b10 b03 b02 b01 b00] */
          "punpcklbw %%mm3, %%mm0\n\t" /* mm0 = [g03 b03 g02 b02 g01 b01 g00 b00] */
          "movq      %%mm0, %%mm2\n\t" /* mm2 = [g03 b03 g02 b02 g01 b01 g00 b00] */
          "movq      %%mm5, %%mm1\n\t" /* mm1 = [b13 b12 b11 b10 b03 b02 b01 b00] */
          "punpcklbw %%mm4, %%mm1\n\t" /* mm1 = [r03 b03 r02 b02 r01 b01 r00 b00] */
          "psrld $8, %%mm1\n\t" /* mm1 = [0 r03 b03 r02 0 r01 b01 r00] */
          "punpcklwd %%mm1, %%mm0\n\t" /* mm0 = [0 r01 g01 b01 b01 r00 g00 b00] */
          "movq      %%mm0, %%mm6\n\t" /* mm6 = [0 r01 g01 b01 b01 r00 g00 b00] */
          "psrlq     $40, %%mm6\n\t" /* mm6 = [0 0 0 0 0 0 r01 g01] */
          "punpckhwd %%mm1, %%mm2\n\t" /* mm2 = [0 r03 g03 b03 b03 r02 g02 b02] */
          "punpcklwd %%mm2, %%mm6\n\t" /* mm6 = [b03 r02 0 0 g02 b02 r01 g01] */
          "punpckldq %%mm6, %%mm0\n\t" /* mm0 = [g02 b02 r01 g01 b01 r00 g00 b00] */
          "movq      %%mm0, (%0)\n\t" /* out 2 + 2/3 pixels (line 1) */
          "psrld     $8, %%mm2\n\t" /* mm1 = [0 0 r03 g03 0 b03 r02 g02] */
          "psrld     $16, %%mm6\n\t" /* mm6 = [0 0 b03 r02 0 0 g02 b02] */
          "punpckhwd %%mm2, %%mm6\n\t" /* mm6 = [0 0 0 0 r03 g03 b03 r02] */
          "movd      %%mm6, 8(%0)\n\t" /* out 1/3 + 1 pixels (line 1) */
          "movq      %%mm5, %%mm0\n\t" /* mm0 = [b13 b12 b11 b10 b03 b02 b01 b00] */
          "punpckhbw %%mm3, %%mm0\n\t" /* mm0 = [g13 b13 g12 b12 g11 b11 g10 b10] */
          "movq      %%mm0, %%mm2\n\t" /* mm2 = [g13 b13 g12 b12 g11 b11 g10 b10] */
          "punpckhbw %%mm4, %%mm5\n\t" /* mm5 = [r13 b13 r12 b12 r11 b11 r10 b10] */
          "psrld     $8, %%mm5\n\t" /* mm5 = [0 r13 b13 r12 0 r11 b11 r10] */
          "punpcklwd %%mm5, %%mm0\n\t" /* mm0 = [0 r11 g11 b11 b11 r10 g10 b10] */
          "movq      %%mm0, %%mm6\n\t" /* mm6 = [0 r11 g11 b11 b11 r10 g10 b10] */
          "psrlq     $40, %%mm6\n\t" /* mm6 = [0 0 0 0 0 0 r11 b11] */
          "punpckhwd %%mm5, %%mm2\n\t" /* mm2 = [0 r13 g13 b13 b13 r12 g12 b12] */
          "punpcklwd %%mm2, %%mm6\n\t" /* mm6 = [b13 r12 0 0 g12 b12 r11 g11] */
          "punpckldq %%mm6, %%mm0\n\t" /* mm0 = [g12 b12 r11 g11 b11 r10 g10 b10] */
          "movq      %%mm0, (%0, %5)\n\t" /* out 2 + 2/3 pixels (line 2) */
          "psrld     $8, %%mm2\n\t" /* mm2 = [0 0 r13 g13 0 b13 r12 g12] */
          "psrld     $16, %%mm6\n\t" /* mm6 = [0 0 b13 r12 0 0 g12 b12] */
          "punpckhwd %%mm2, %%mm6\n\t" /* mm6 = [0 0 0 0 r13 g13 b13 r12] */
          "movd      %%mm6, 8(%0, %5)\n\t" /* out 1/3 + 1 pixels (line 2) */
          "addl      $12, %0\n\t"
          "addl      $4, %1\n\t"
          "addl      $2, %2\n\t"
          "addl      $2, %3\n\t"
          "decl      %6\n\t" /* --line */
          "jnz       2f\n\t"
          "movl      %4, %6\n\t" /* if (!line) { */
          "shrl      $2, %6\n\t" /* line = frame->width.cur / 4 */
          "addl      %5, %0\n\t" /* dst += pitch */
          "addl      %4, %1\n\t" /* srcy += frame->width.cur } */
          "2:\n\t"
          "decl      %7\n\t"
          "jnz       1b\n\t"
          "emms"

          :: "r" (dst), "r" (srcy), "r" (srcu), "r" (srcv),
             "q" (frame->width.cur), "q" (pitch), "m" (line), "m" (n)
          : "memory");
}


static
DFB_PFUNCTION( yv12, rgb32 )
{
     uint8_t *srcy = frame->vo_frame.base[0];
     uint8_t *srcu = frame->vo_frame.base[1];
     uint8_t *srcv = frame->vo_frame.base[2];
     int      n    = (frame->width.cur * frame->height.cur) >> 3;
     int      line = frame->width.cur >> 2;

     __asm__ __volatile__(

          "pxor      %%mm7, %%mm7\n\t"
          ".align 16\n"
          "1:\n\t"
          "movd      (%1), %%mm3\n\t" /* mm3 = [0 0 0 0 y03 y02 y01 y00] */
          "punpcklbw %%mm7, %%mm3\n\t" /* mm3 = [0 y03 0 y02 0 y01 0 y00] */
          "paddw     (mmx_table + 32), %%mm3\n\t" /* y + brightness */
          "psllw     $2, %%mm3\n\t" /* y << 2 */
          "pmulhw    (mmx_table + 40), %%mm3\n\t" /* y * contrast */
          "movd      (%1, %4), %%mm2\n\t" /* mm2 = [0 0 0 0 y13 y12 y11 y10] */
          "punpcklbw %%mm7, %%mm2\n\t" /* mm2 = [0 y13 0 y12 0 y11 0 y10] */
          "paddw     (mmx_table + 32), %%mm2\n\t" /* y + brightness */
          "psllw     $2, %%mm2\n\t" /* y << 2 */
          "pmulhw    (mmx_table + 40), %%mm2\n\t" /* y * contrast */
          "movd      (%2), %%mm1\n\t" /* mm1 = [0 0 0 0 u3 u2 u1 u0] */
          "punpcklbw %%mm1, %%mm1\n\t" /* mm1 = [u3 u3 u2 u2 u1 u1 u0 u0] */
          "punpcklbw %%mm7, %%mm1\n\t" /* mm1 = [0 u1 0 u1 0 u0 0 u0] */
          "psubw     (chroma_sub), %%mm1\n\t" /* u -= 128 */
          "psllw     $3, %%mm1\n\t" /* u << 3 */
          "movd      (%3), %%mm0\n\t" /* mm0 = [0 0 0 0 v3 v2 v1 v0] */
          "punpcklbw %%mm0, %%mm0\n\t" /* mm0 = [v3 v3 v2 v2 v1 v1 v0 v0] */
          "punpcklbw %%mm7, %%mm0\n\t" /* mm0 = [0 v1 0 v1 0 v0 0 v0] */
          "psubw     (chroma_sub), %%mm0\n\t" /* v -= 128 */
          "psllw     $3, %%mm0\n\t" /* v << 3 */
          "movq      %%mm0, %%mm4\n\t" /* mm4 = [0 v1 0 v1 0 v0 0 v0] */
          "pmulhw    (mmx_table), %%mm4\n\t" /* chroma_r*/
          "movq      %%mm4, %%mm5\n\t"
          "paddw     %%mm3, %%mm4\n\t" /* mm4 = [r03 | r02 | r01 | r00] */
          "paddw     %%mm2, %%mm5\n\t" /* mm5 = [r13 | r12 | r11 | r10] */
          "packuswb  %%mm5, %%mm4\n\t" /* mm4 = [r13 r12 r11 r10 r03 r02 r01 r00] */
          "movq      %%mm1, %%mm5\n\t" /* mm5 = [0 u1 0 u1 0 u0 0 u0] */
          "pmulhw    (mmx_table + 24), %%mm5\n\t" /* chroma_b */
          "movq      %%mm5, %%mm6\n\t"
          "paddw     %%mm3, %%mm5\n\t" /* mm5 = [b03 | b02 | b01 | b00] */
          "paddw     %%mm2, %%mm6\n\t" /* mm6 = [b13 | b12 | b11 | b00] */
          "packuswb  %%mm6, %%mm5\n\t" /* mm5 = [b13 b12 b11 b10 b03 b02 b01 b00] */
          "pmulhw    (mmx_table + 8), %%mm0\n\t"
          "pmulhw    (mmx_table + 16), %%mm1\n\t"
          "paddw     %%mm1, %%mm0\n\t" /* chorma_g */
          "psubw     %%mm0, %%mm3\n\t" /* mm3 = [g03 | g02 | g01 | g00] */
          "psubw     %%mm0, %%mm2\n\t" /* mm2 = [g13 | g12 | g11 | g10] */
          "packuswb  %%mm2, %%mm3\n\t" /* mm3 = [g13 g12 g11 g10 g03 g02 g01 g00] */
          "movq      %%mm4, %%mm0\n\t" /* mm0 = [r13 r12 r11 r10 r03 r02 r01 r00] */
          "punpcklbw %%mm7, %%mm0\n\t" /* mm0 = [0 r03 0 r02 0 r01 0 r00] */
          "movq      %%mm5, %%mm1\n\t" /* mm1 = [b13 b12 b11 b10 b03 b02 b01 b00] */
          "punpcklbw %%mm3, %%mm1\n\t" /* mm1 = [g03 b03 g02 b02 g01 b01 g00 b00] */
          "movq      %%mm1, %%mm2\n\t" /* mm2 = [g03 b03 g02 b02 g01 b01 g00 b00] */
          "punpcklwd %%mm0, %%mm2\n\t" /* mm2 = [0 r01 g01 b01 0 r00 g00 b00] */
          "movq      %%mm2, (%0)\n\t" /* out first 2 pixels (line 1) */
          "punpckhwd %%mm0, %%mm1\n\t" /* mm1 = [0 r03 g03 b03 0 r02 g02 b02] */
          "movq      %%mm1, 8(%0)\n\t" /* out second 2 pixels (line 1) */
          "punpckhbw %%mm7, %%mm4\n\t" /* mm4 = [0 r13 0 r12 0 r11 0 r10] */
          "punpckhbw %%mm3, %%mm5\n\t" /* mm5 = [g13 b13 g12 b12 g11 b11 g10 b10] */
          "movq      %%mm5, %%mm0\n\t" /* mm0 = [g13 b13 g12 b12 g11 b11 g10 b10] */
          "punpcklwd %%mm4, %%mm0\n\t" /* mm0 = [0 r11 g11 b11 0 r10 g10 b10] */
          "movq      %%mm0, (%0, %5)\n\t" /* out first 2 pixels (line 2) */
          "punpckhwd %%mm4, %%mm5\n\t" /* mm5 = [0 r13 g13 b13 0 r12 g12 b12] */
          "movq      %%mm5, 8(%0, %5)\n\t" /* out second 2 pixels (line 2) */
          "addl      $16, %0\n\t"
          "addl      $4, %1\n\t"
          "addl      $2, %2\n\t"
          "addl      $2, %3\n\t"
          "decl      %6\n\t" /* --line */
          "jnz       2f\n\t"
          "movl      %4, %6\n\t" /* if (!line) { */
          "shrl      $2, %6\n\t" /* line = frame->width.cur / 4 */
          "addl      %5, %0\n\t" /* dst += pitch */
          "addl      %4, %1\n\t" /* srcy += frame->width.cur } */
          "2:\n\t"
          "decl      %7\n\t"
          "jnz       1b\n\t"
          "emms"

          :: "r" (dst), "r" (srcy), "r" (srcu), "r" (srcv),
             "q" (frame->width.cur), "q" (pitch), "m" (line), "m" (n)
          : "memory");
}


static
DFB_PFUNCTION( yv12, argb )
{
     uint8_t *srcy = frame->vo_frame.base[0];
     uint8_t *srcu = frame->vo_frame.base[1];
     uint8_t *srcv = frame->vo_frame.base[2];
     int      n    = (frame->width.cur * frame->height.cur) >> 3;
     int      line = frame->width.cur >> 2;

     __asm__ __volatile__(

          "pxor      %%mm7, %%mm7\n\t"
          ".align 16\n"
          "1:\n\t"
          "movd      (%1), %%mm3\n\t" /* mm3 = [0 0 0 0 y03 y02 y01 y00] */
          "punpcklbw %%mm7, %%mm3\n\t" /* mm3 = [0 y03 0 y02 0 y01 0 y00] */
          "paddw     (mmx_table + 32), %%mm3\n\t" /* y + brightness */
          "psllw     $2, %%mm3\n\t" /* y << 2 */
          "pmulhw    (mmx_table + 40), %%mm3\n\t" /* y * contrast */
          "movd      (%1, %4), %%mm2\n\t" /* mm2 = [0 0 0 0 y13 y12 y11 y10] */
          "punpcklbw %%mm7, %%mm2\n\t" /* mm2 = [0 y13 0 y12 0 y11 0 y10] */
          "paddw     (mmx_table + 32), %%mm2\n\t" /* y + brightness */
          "psllw     $2, %%mm2\n\t" /* y << 2 */
          "pmulhw    (mmx_table + 40), %%mm2\n\t" /* y * contrast */
          "movd      (%2), %%mm1\n\t" /* mm1 = [0 0 0 0 u3 u2 u1 u0] */
          "punpcklbw %%mm1, %%mm1\n\t" /* mm1 = [u3 u3 u2 u2 u1 u1 u0 u0] */
          "punpcklbw %%mm7, %%mm1\n\t" /* mm1 = [0 u1 0 u1 0 u0 0 u0] */
          "psubw     (chroma_sub), %%mm1\n\t" /* u -= 128 */
          "psllw     $3, %%mm1\n\t" /* u << 3 */
          "movd      (%3), %%mm0\n\t" /* mm0 = [0 0 0 0 v3 v2 v1 v0] */
          "punpcklbw %%mm0, %%mm0\n\t" /* mm0 = [v3 v3 v2 v2 v1 v1 v0 v0] */
          "punpcklbw %%mm7, %%mm0\n\t" /* mm0 = [0 v1 0 v1 0 v0 0 v0] */
          "psubw     (chroma_sub), %%mm0\n\t" /* v -= 128 */
          "psllw     $3, %%mm0\n\t" /* v << 3 */
          "movq      %%mm0, %%mm4\n\t" /* mm4 = [0 v1 0 v1 0 v0 0 v0] */
          "pmulhw    (mmx_table), %%mm4\n\t" /* chroma_r*/
          "movq      %%mm4, %%mm5\n\t"
          "paddw     %%mm3, %%mm4\n\t" /* mm4 = [r03 | r02 | r01 | r00] */
          "paddw     %%mm2, %%mm5\n\t" /* mm5 = [r13 | r12 | r11 | r10] */
          "packuswb  %%mm5, %%mm4\n\t" /* mm4 = [r13 r12 r11 r10 r03 r02 r01 r00] */
          "movq      %%mm1, %%mm5\n\t" /* mm5 = [0 u1 0 u1 0 u0 0 u0] */
          "pmulhw    (mmx_table + 24), %%mm5\n\t" /* chroma_b */
          "movq      %%mm5, %%mm6\n\t"
          "paddw     %%mm3, %%mm5\n\t" /* mm5 = [b03 | b02 | b01 | b00] */
          "paddw     %%mm2, %%mm6\n\t" /* mm6 = [b13 | b12 | b11 | b00] */
          "packuswb  %%mm6, %%mm5\n\t" /* mm5 = [b13 b12 b11 b10 b03 b02 b01 b00] */
          "pmulhw    (mmx_table + 8), %%mm0\n\t"
          "pmulhw    (mmx_table + 16), %%mm1\n\t"
          "paddw     %%mm1, %%mm0\n\t" /* chorma_g */
          "psubw     %%mm0, %%mm3\n\t" /* mm3 = [g03 | g02 | g01 | g00] */
          "psubw     %%mm0, %%mm2\n\t" /* mm2 = [g13 | g12 | g11 | g10] */
          "packuswb  %%mm2, %%mm3\n\t" /* mm3 = [g13 g12 g11 g10 g03 g02 g01 g00] */
          "movq      %%mm4, %%mm0\n\t" /* mm0 = [r13 r12 r11 r10 r03 r02 r01 r00] */
          "punpcklbw %%mm7, %%mm0\n\t" /* mm0 = [0 r03 0 r02 0 r01 0 r00] */
          "por       (amask), %%mm0\n\t" /* mm0 = [ff r03 ff r02 ff r01 ff r00] */
          "movq      %%mm5, %%mm1\n\t" /* mm1 = [b13 b12 b11 b10 b03 b02 b01 b00] */
          "punpcklbw %%mm3, %%mm1\n\t" /* mm1 = [g03 b03 g02 b02 g01 b01 g00 b00] */
          "movq      %%mm1, %%mm2\n\t" /* mm2 = [g03 b03 g02 b02 g01 b01 g00 b00] */
          "punpcklwd %%mm0, %%mm2\n\t" /* mm2 = [ff r01 g01 b01 ff r00 g00 b00] */
          "movq      %%mm2, (%0)\n\t" /* out first 2 pixels (line 1) */
          "punpckhwd %%mm0, %%mm1\n\t" /* mm1 = [ff r03 g03 b03 ff r02 g02 b02] */
          "movq      %%mm1, 8(%0)\n\t" /* out second 2 pixels (line 1) */
          "punpckhbw %%mm7, %%mm4\n\t" /* mm4 = [0 r13 0 r12 0 r11 0 r10] */
          "por       (amask), %%mm4\n\t" /* mm4 = [ff r13 ff r12 ff r11 ff r10 */
          "punpckhbw %%mm3, %%mm5\n\t" /* mm5 = [g13 b13 g12 b12 g11 b11 g10 b10] */
          "movq      %%mm5, %%mm0\n\t" /* mm0 = [g13 b13 g12 b12 g11 b11 g10 b10] */
          "punpcklwd %%mm4, %%mm0\n\t" /* mm0 = [ff r11 g11 b11 ff r10 g10 b10] */
          "movq      %%mm0, (%0, %5)\n\t" /* out first 2 pixels (line 2) */
          "punpckhwd %%mm4, %%mm5\n\t" /* mm5 = [ff r13 g13 b13 ff r12 g12 b12] */
          "movq      %%mm5, 8(%0, %5)\n\t" /* out second 2 pixels (line 2) */
          "addl      $16, %0\n\t"
          "addl      $4, %1\n\t"
          "addl      $2, %2\n\t"
          "addl      $2, %3\n\t"
          "decl      %6\n\t" /* --line */
          "jnz       2f\n\t"
          "movl      %4, %6\n\t" /* if (!line) { */
          "shrl      $2, %6\n\t" /* line = frame->width.cur / 4 */
          "addl      %5, %0\n\t" /* dst += pitch */
          "addl      %4, %1\n\t" /* srcy += frame->width.cur } */
          "2:\n\t"
          "decl      %7\n\t"
          "jnz       1b\n\t"
          "emms"

          :: "r" (dst), "r" (srcy), "r" (srcu), "r" (srcv),
             "q" (frame->width.cur), "q" (pitch), "m" (line), "m" (n)
          : "memory");
}


/******************************** End of YV12 callbacks *********************************/


static const DVProcFunc PACCEL[2][DFB_NUM_PIXELFORMATS] =
{
     /* YUY2 */
     {
          DFB_PFUNCTION_NAME( PACCEL, yuy2, rgb15 ),
          DFB_PFUNCTION_NAME( PACCEL, yuy2, rgb16 ),
          DFB_PFUNCTION_NAME( PACCEL, yuy2, rgb24 ),
          DFB_PFUNCTION_NAME( PACCEL, yuy2, rgb32 ),
          DFB_PFUNCTION_NAME( PACCEL, yuy2, argb ),
          NULL,
          DFB_PFUNCTION_NAME( PACCEL, yuy2, yuy2 ),
          DFB_PFUNCTION_NAME( PACCEL, yuy2, rgb8 ),
          DFB_PFUNCTION_NAME( PACCEL, yuy2, uyvy ),
          DFB_PFUNCTION_NAME( PACCEL, yuy2, yv12 ),
          DFB_PFUNCTION_NAME( PACCEL, yuy2, yv12 ),
          NULL,
          NULL,
          DFB_PFUNCTION_NAME( PACCEL, yuy2, rgb32 )
     },
     /* YV12 */
     {
          DFB_PFUNCTION_NAME( PACCEL, yv12, rgb15 ),
          DFB_PFUNCTION_NAME( PACCEL, yv12, rgb16 ),
          DFB_PFUNCTION_NAME( PACCEL, yv12, rgb24 ),
          DFB_PFUNCTION_NAME( PACCEL, yv12, rgb32 ),
          DFB_PFUNCTION_NAME( PACCEL, yv12, argb ),
          NULL,
          DFB_PFUNCTION_NAME( PACCEL, yv12, yuy2 ),
          DFB_PFUNCTION_NAME( PACCEL, yv12, rgb8 ),
          DFB_PFUNCTION_NAME( PACCEL, yv12, uyvy ),
          DFB_PFUNCTION_NAME( PACCEL, yv12, yv12 ),
          DFB_PFUNCTION_NAME( PACCEL, yv12, yv12 ),
          NULL,
          NULL,
          DFB_PFUNCTION_NAME( PACCEL, yv12, rgb32 )
     }
};


#undef PACCEL
