/*
 * Copyright (C) 2004 Claudio "KLaN" Ciccani <klan82@cheapnet.it>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *  written by Claudio Ciccani <klan82@cheapnet.it> with the contribute
 *  of Kristof Pelckmans <kristof.pelckmans@antwerpen.be>
 *
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/types.h>

#include "df_xine.h"
#include "keymap.h"



static DFXCore core;



void
dfx_exit( int status )
{
     dfx_control_release( &core);

     dfx_context_release( &core );

     if (core.bg)
          core.bg->Release( core.bg );

     if (core.primary)
          core.primary->Release( core.primary );

     if (core.buffer)
          core.buffer->Release( core.buffer );

     if (core.dfb)
     {
          core.dfb->SetCooperativeLevel( core.dfb, DFSCL_NORMAL);
          core.dfb->Release( core.dfb );
     }

     dfx_playlist_free( &core );

     if (core.verbosity)
          fprintf(stderr, THIS ": exiting with status %i\n", status );

     exit( status );
}


static void
dfx_usage( void )
{
     int i;

     puts( "\nUsage: " THIS " [options] MRLs...\n" );
     puts( "    -h  , --help          show this help\n" );
     puts( "    -v N, --verbosity=N   set verbosity level to N" );
     puts( "                          (minimum: 0, maximum: 2, default: 1)\n" );
     puts( "    -l N, --layer=N       use layer N for video playback" );
     puts( "                          (default: auto)\n" );
     puts( "    -b S, --buffermode=S  use buffermode S (single, double or triple)" );
     puts( "                          (default: double)\n" );
     puts( "    -s  , --sync          wait for vertical retrace" );
     puts( "                          (default: don't wait)\n" );
     puts( "    -r N, --repeat=N      repeat playlist N times" );
     puts( "                          (default: 1)\n" );
     puts( "    -R  , --recursive     repeat playlist infinite times\n" );
     puts( "          --stdctl        turn on controlling xine over STDIN\n");
     puts( "    Use:" );

     for (i = 0; i < (sizeof( kmap ) / sizeof( kmap[0] )); i++)
     {
          if (kmap[i].event)
               continue;

          if (kmap[i].subkey)
               printf( "        %-4s + %-6s\tto %s\n",
                    kmap[i].name[0], kmap[i].name[1], kmap[i].help );
          else
               printf( "        %-4s%9c\tto %s\n",
                    kmap[i].name[0], ' ',  kmap[i].help );
     }

     puts( "\n    For DVD navigation use:" );

     for (i = 0; i < (sizeof( kmap ) / sizeof( kmap[0] )); i++)
     {
          if (!kmap[i].event)
               continue;

          if (kmap[i].subkey)
               printf( "        %-4s + %-6s\tto %s\n",
                    kmap[i].name[0], kmap[i].name[1], kmap[i].help );
          else
               printf( "        %-4s%9c\tto %s\n",
                    kmap[i].name[0], ' ',  kmap[i].help );
     }

     putchar( '\n' );

     exit( EXIT_FAILURE );
}


static void
dfx_init( DFXCore *this, int *argn, char ***argp )
{
     char **argv;
     int    argc;
     int    i;

     if (*argn < 2)
          dfx_usage();

     memset( this, 0, sizeof( DFXCore ) );

     this->verbosity      =  1;
     this->ctx.lid        = -1;
     this->ctx.buffermode = DLBM_BACKVIDEO;
     this->ctx.gmode      = DGM_VIDEO;
     this->media.repeat   =  1;
     this->stdctl         =  0;

     DirectFBInit( argn, argp );

     for (argv = *argp, argc = *argn, i = 1; i < argc; i++)
     {
          char *option = argv[i];

          if (!option || !*option)
               continue;

          if (*option == '-')
          {
               option++;

               if (*option == 'h' || !strcmp( option, "-help" ))
                    dfx_usage();

               if (*option == 'v')
               {
                    if (isdigit( *(option + 1) ))
                         option++;
                    else
                    if (argv[i + 1] && isdigit( *argv[i + 1] ))
                         option = argv[++i];
                    else
                         continue;
                    
                    this->verbosity = *option - '0';
               } else
               if (!strncmp( option, "-verbosity=", 11 ))
               {
                    option += 11;
                    if (isdigit( *option ))
                         this->verbosity = *option - '0';
               } else
               if (*option == 'l')
               {
                    if (isdigit( *(option + 1) ))
                         option++;
                    else
                    if (argv[i + 1] && isdigit( *argv[i + 1] ))
                         option = argv[++i];
                    else
                         continue;

                    this->ctx.lid = strtol( option, NULL, 10 );
               } else
               if (!strncmp( option, "-layer=", 7 ))
               {
                    option += 7;
                    if (isdigit( *option ))
                         this->ctx.lid = strtol( option, NULL, 10 );
               } else
               if (*option == 'b')
               {
                    if (argv[i + 1] && *argv[i + 1] && *argv[i + 1] != '-')
                         option = argv[++i];
                    else
                         continue;
                    
                    if (!strcasecmp( option, "single" ))
                         this->ctx.buffermode = DLBM_FRONTONLY;
                    else
                    if (!strcasecmp( option, "double" ))
                         this->ctx.buffermode = DLBM_BACKVIDEO;
                    else
                    if (!strcasecmp( option, "triple" ))
                         this->ctx.buffermode = DLBM_TRIPLE;
               } else
               if (!strncmp( option, "-buffermode=", 12 ))
               {
                    option += 12;
                    if (!*option)
                         continue;

                    if (!strcasecmp( option, "single" ))
                         this->ctx.buffermode = DLBM_FRONTONLY;
                    else
                    if (!strcasecmp( option, "double" ))
                         this->ctx.buffermode = DLBM_BACKVIDEO;
                    else
                    if (!strcasecmp( option, "triple" ))
                         this->ctx.buffermode = DLBM_TRIPLE;
               } else
               if (!strcmp( option, "s" ) || !strcmp( option, "-sync" ))
               {
                    this->ctx.flipflags = DSFLIP_WAITFORSYNC;
               } else                    
               if (!strcmp( option, "R" ) || !strcmp( option, "-recursive" ))
               {
                    this->media.repeat = -1;
               } else
               if (*option == 'r')
               {
                    if (isdigit( *(option + 1) ))
                         option++;
                    else
                    if (argv[i + 1] && isdigit( *argv[i + 1] ))
                         option = argv[++i];
                    else
                         continue;
                    
                    this->media.repeat = strtol( option, NULL, 10 );
               } else
               if (!strncmp( option, "-repeat=", 8 ))
               {
                    option += 8;
                    if (isdigit( *option ))
                         this->media.repeat = strtol( option, NULL, 10 );
               } else
	           if (!strncmp( option, "-stdctl" ,7))
		           this->stdctl = 1;
          } else
               dfx_playlist_append( this, option );
     }

     /* is playlist empty ? */
     if (!this->media.first)
          FATAL( "no MRL selected, playlist is empty\n" );
}



static inline int
kmap_lookup_key( int key )
{
     int i;

     for (i = 0; i < (sizeof( kmap ) / sizeof( kmap[0] )); i++)
     {
          if (kmap[i].key == key)
               return i;
     }

     return -1;
}


static inline int
kmap_lookup_keycomb( int key, int subkey )
{
     int i;

     for (i = 0; i < (sizeof( kmap ) / sizeof( kmap[0] )); i++)
     {
          if (kmap[i].key == key && kmap[i].subkey == subkey)
               return i;
     }

     return -1;
}


static int
dfx_handle_keypress( DFXCore *this, int key, int symbol )
{
     int id;
     int subkey;
     int magic;

     if (key == -1)
     {
          key = tolower( symbol );
          id  = kmap_lookup_key( key );

          if (id < 0)
               return -1;

          if (!kmap[id].subkey)
          {
               magic = DFX_MAGIC( kmap[id].event, 0,
                            kmap[id].actid, kmap[id].param );
               key   = -1;
          }
          else 
               magic = DFX_MAGIC( kmap[id].event, DAN_ACQUIRE,
                            kmap[id].actid, kmap[id].param );
               
          dfx_control_action( this, magic );
     } else
     {
          subkey = tolower( symbol );
          id     = kmap_lookup_keycomb( key, subkey );

          if (id < 0)
               return key;

          magic = DFX_MAGIC( kmap[id].event, 0,
                       kmap[id].actid, kmap[id].param );
          dfx_control_action( this, magic );
     }

     return key;
}


static int
dfx_handle_keyrelease( DFXCore *this, int key, int symbol )
{
     int id, magic;

     if (key == tolower( symbol ))
     {
          id = kmap_lookup_key( key );
          
          magic = DFX_MAGIC( kmap[id].event, DAN_RELEASE,
                       kmap[id].actid, kmap[id].param );
          dfx_control_action( this, magic );

          key = -1;
     }

     return key;
}


static void
dfx_event_loop( DFXCore *this )
{
     IDirectFBEventBuffer *buffer;

     DFBCHECK(this->dfb->CreateInputEventBuffer( this->dfb,
                                    DICAPS_KEYS, 0, &buffer ));
     this->buffer = buffer;

     while (!this->quit)
     {
          DFBInputEvent evt;
          static int    key = -1;

          buffer->WaitForEventWithTimeout( buffer, 0, 10 );

          while (buffer->GetEvent( buffer, DFB_EVENT( &evt) ) == DFB_OK)
          {
               switch (evt.type)
               {
                    case DIET_KEYPRESS:
                         key = dfx_handle_keypress( this,
                                             key, evt.key_symbol );
                    break;

                    case DIET_KEYRELEASE:
                         key = dfx_handle_keyrelease( this,
                                             key, evt.key_symbol );
                    break;

                    default:
                    break;
               }
          }
     }
}


/*
 * Chris Griffiths <chris@y-fronts.com>:
 *
 *   Alternate main loop - reads commands from stdin, uses keymap.h to determine
 *   required action and requests it. Used to run xine from within 
 *   freevo (www.freevo.org).
 *                 
 *   Clears OSD automatically OSD_CLEAR_TIME/2 secs after last action
 */

#define OSD_CLEAR_TIME 7

static inline int
dfx_stdctl_get_action_magic( char action_name[255] )
{
     int id;

     for (id = 0; id < (sizeof( kmap ) / sizeof( kmap[0] )); id++)
     {
          if (kmap[id].stdctl && !strcasecmp( kmap[id].stdctl, action_name ))
               return DFX_MAGIC( kmap[id].event, 0, kmap[id].actid, kmap[id].param );
     }
     return -1;
}


static void
dfx_stdctl_loop( DFXCore *this ) 
{
     char           c[255];
     int            len;
     int            key_magic;
     fd_set         readset;
     struct timeval timeout;
     int            osd_clear_delay=-1;

     while (!this->quit)
     {
          osd_clear_delay--;
          if (osd_clear_delay == 0)
          {
               dfx_display_hide( this, 1 );
               osd_clear_delay = -1;
          }

          timeout.tv_sec  =  0;
          timeout.tv_usec =  500000;

          FD_ZERO( &readset );
          FD_SET( STDIN_FILENO, &readset );
          select( STDIN_FILENO + 1, &readset, NULL, NULL, &timeout );

          if (FD_ISSET( STDIN_FILENO, &readset ))
          {
               len = read( STDIN_FILENO, &c, 255 );
               if (len > 0)
               {
                    c[len-1] = '\0';
                    key_magic = dfx_stdctl_get_action_magic( c );
                    if (key_magic != -1)
                    {
                         dfx_control_action( this, key_magic );
                         osd_clear_delay=OSD_CLEAR_TIME;
                     }
               }
          }
     }
}


int
main( int argc, char *argv[] )
{
     DFXCore               *this = &core;
     DFBSurfaceDescription  desc =
     {
          .flags = DSDESC_CAPS,
          .caps  = DSCAPS_PRIMARY
     };

     dfx_init( this, &argc, &argv );

     DFBCHECK(DirectFBCreate( &this->dfb ));
     this->dfb->SetCooperativeLevel( this->dfb, DFSCL_FULLSCREEN );
     SAY( "here we go!\n" );

     DFBCHECK(this->dfb->GetDisplayLayer( this->dfb, DLID_PRIMARY,
                               &this->primary ));
     DFBCHECK(this->dfb->CreateSurface( this->dfb, &desc, &this->bg ));
     this->bg->Clear( this->bg, 0, 0, 0, 0xff );

     dfx_context_create( this );
     dfx_control_start( this );

     if (this->stdctl)
               dfx_stdctl_loop( this );
     else
               dfx_event_loop( this );

     dfx_exit( this->quit - 1 );

     return 0;
}

